from ._add_edges import add_edges
from ._floyd_warshall import floyd_warshall
from ._matrix import adj_matrix
from ._matrix import adj_matrix_from_edges
from ._matrix import graph_matrix
from ._matrix import in_degree
from ._matrix import in_degree_matrix_from_edges
from ._matrix import out_degree
from ._matrix import out_degree_matrix_from_edges
from ._sparse import to_sparse_coo_matrix
from ._traversal import bfs_nodes
from ._traversal import induce
from ._traversal import neighbors
from ._traversal import tensor_induce
from ._utils import get_edge_dict

__all__ = [
    "add_edges",
    "floyd_warshall",
    "adj_matrix",
    "adj_matrix_from_edges",
    "graph_matrix",
    "in_degree",
    "in_degree_matrix_from_edges",
    "out_degree",
    "out_degree_matrix_from_edges",
    "to_sparse_coo_matrix",
    "bfs_nodes",
    "induce",
    "neighbors",
    "tensor_induce",
    "get_edge_dict",
]



r"""
Transformation classes for :class:`caldera.data.GraphData` and :class:`caldera.data.GraphBatch`.

.. autosummary::
    :toctree: generated/

   Shuffle
   Reverse
   Undirected
   FullyConnected
   RandomEdgeMask
   RandomNodeMask
   RandomHop
"""
from ._add_self_loops import AddSelfLoops
from ._compose import Compose
from ._fully_connected import FullyConnected
from ._random_edge_mask import RandomEdgeMask
from ._random_hop import RandomHop
from ._random_node_mask import RandomNodeMask
from ._reverse import Reverse
from ._shuffle import Shuffle
from ._undirected import Undirected

__all__ = [
    "AddSelfLoops",
    "Compose",
    "FullyConnected",
    "RandomHop",
    "RandomEdgeMask",
    "RandomNodeMask",
    "Reverse",
    "Shuffle",
    "Undirected",
]



r"""
This module provide objects for representing graphs as `torch` tensors.

.. autosummary::
    :toctree: generated/

    GraphBatch
    GraphData
    GraphTuple
    GraphDataLoader
"""
from ._graph_batch import GraphBatch
from ._graph_data import GraphData
from ._graph_tuple import GraphTuple
from ._loader import GraphDataLoader

__all__ = [
    "GraphBatch",
    "GraphData",
    "GraphTuple",
    "GraphDataLoader",
]


def method():
    """This is bullshit."""


this_is_my_attribute = "a tribute"



# import random
#
# import networkx as nx
#
#
# # TODO: add long traversal
# def random_graph(n, e=None, d=None):
#     """Create a random graph."""
#     n = random.randint(*n)
#     if e is None:
#         d = random.random() * d[1] + d[0]
#         e = int((d * n * (n - 1)) / 2)
#     else:
#         e = random.randint(*e)
#     e = max(1, e)
#     return nx.generators.dense_gnm_random_graph(n, e)
#
#
# def annotate_shortest_path(g):
#     source, target = np.random.choice(list(g.nodes), size=(2,))
#
#     g.nodes[source]["source"] = True
#     g.nodes[target]["target"] = True
#     try:
#         traversal = nx.shortest_path(g, source=nodes[0], target=nodes[1])
#     except nx.NetworkXNoPath:
#         traversal = []
#
#     for n in traversal:
#         g.nodes[n]["shortest_path"] = False
#     for n, ndata in g.nodes(data=True):
#         ndata["shortest_path"] = target
#
#     for n1, n2, edata in g.edges(data=True):
#         edata["shortest_path"] = False
#
#     for n1, n2 in nx.utils.pairwise(traversal):
#         g[n1][n2]["shortest_path"] = True
#
#
# def cat_property(from_key, to_key):
#     pass
#
#
# def from_property_graph_to_graph_data():
#     pass



r"""
Design (:mod:`dasi.design`)
=============================

.. currentmodule:: caldera.data._graph_data

This module provide DNA assembly functionality for DASi.
"""
from __future__ import annotations

import functools
from functools import reduce
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import networkx as nx
import numpy as np
import torch

from caldera.utils import long_isin
from caldera.utils import reindex_tensor
from caldera.utils import same_storage
from caldera.utils.nx import nx_is_directed
from caldera.utils.nx.types import DirectedGraph


def np_or_tensor_size(arr: Union[torch.tensor, np.ndarray]) -> int:
    if issubclass(arr.__class__, torch.Tensor):
        return arr.nelement()
    elif issubclass(arr.__class__, np.ndarray):
        return arr.size
    else:
        raise ValueError("Must be a {} or {}".format(np.ndarray, torch.Tensor))


# TODO: there should be a super class, TorchComposition, with apply methods etc.
# TODO: support n dim tensors
# TODO: implicit support for torch.Tensor
# TODO: handle empty features and targets
class GraphData:
    """Data representing a single graph."""

    __slots__ = ["x", "e", "g", "edges"]
    _differentiable = ["x", "e", "g"]

    def __init__(
        self,
        node_attr: torch.FloatTensor,
        edge_attr: torch.FloatTensor,
        global_attr: torch.FloatTensor,
        edges: torch.LongTensor,
        requires_grad: Optional[bool] = None,
    ):
        """Blank.

        :param node_attr:
        :param edge_attr:
        :param global_attr:
        :param edges:
        :param requires_grad:
        """
        self.x = node_attr
        self.e = edge_attr
        self.g = global_attr
        self.edges = edges
        GraphData.debug(self)
        if requires_grad is not None:
            self.requires_grad = requires_grad

    def debug(self):
        if (
            self.edges.shape[0]
            and self.edges.shape[1]
            and self.edges.max() >= self.x.shape[0]
        ):
            raise RuntimeError(
                "Edge coordinate {} is greater than number of nodes {}".format(
                    self.edges.max(), self.x.shape[0]
                )
            )

        if self.edges.shape[0]:
            n_edges = self.edges.shape[1]
        else:
            n_edges = 0
        if not n_edges == self.e.shape[0]:
            raise RuntimeError(
                "Number of edges {} must match number of edge attributes {}".format(
                    n_edges, self.e.shape[0]
                )
            )

        if not self.edges.dtype == torch.long:
            raise RuntimeError(
                "Wrong tensor type. `edges` must be dtype={} not {}".format(
                    torch.long, self.edges.dtype
                )
            )

        if not self.edges.shape[0] == 2:
            raise RuntimeError("Edges must be a tensor of shape `[2, num_edges]`")

        if not self.x.ndim == 2:
            raise RuntimeError("Node attr must have 2 dimensions")

        if not self.g.ndim == 2:
            raise RuntimeError("Global attr must have 2 dimensions")

        if not self.e.ndim == 2:
            raise RuntimeError("Edge attr must have 2 dimensions")

        if not self.edges.ndim == 2:
            raise RuntimeError("Edges must have 2 dimensions")

    def _apply(
        self,
        func,
        new_inst: bool,
        args: Tuple[Any, ...] = tuple(),
        kwargs: Dict = None,
        keys: Optional[Tuple[str]] = None,
    ) -> "GraphData":
        """Applies the function to the graph. Be mindful of what the function
        is doing.

        Ask the following before using this function:

        1. Is it applying the function in place? copying it?
        2. Is the function recorded in the computation graph? Will you need to detach it?

        :param func: function to apply
        :param new_inst: Whether to create a new GraphData instance. Values will be copied (and detached by default).
        :param detatch_new_inst: Detaches copied vectors from the computation graph. This should almost always be left
            `True`.
        :param args:
        :param kwargs:
        :return:
        """
        if keys is None:
            keys = [k for k, _ in self._tensors]
        if kwargs is None:
            kwargs = {}
        init_args = []
        for f in keys:
            old_val = getattr(self, f)
            val = func(old_val, *args, **kwargs)
            if new_inst:
                init_args.append(val)
        if new_inst:
            return self.__class__(*init_args)
        return self

    # TODO: finish clone, copy, apply, etc.
    def apply(self, func, *args, keys: Optional[Tuple[str]] = None, **kwargs):
        """Applies the function to the data, creating a new instance of
        GraphData."""
        return self._apply(func, new_inst=True, args=args, kwargs=kwargs, keys=keys)

    def apply_(self, func, *args, keys: Optional[Tuple[str]] = None, **kwargs):
        """Applies the function in place to the data, wihout creating a new
        instance of GraphData."""
        return self._apply(func, new_inst=False, args=args, kwargs=kwargs, keys=keys)

    def detach(self):
        return self.apply(lambda x: x.detach())

    def to(self, device: str, *args, **kwargs):
        return self.apply(lambda x: x.to(device, *args, **kwargs))

    @property
    def _tensors(self):
        for k in self.__slots__:
            if not k.startswith("__"):
                v = getattr(self, k)
                if torch.is_tensor(v):
                    yield k, v

    def share_storage(
        self, other: GraphData, return_dict: Optional[bool] = False
    ) -> Union[Dict[str, bool], bool]:
        """Check if this data shares storage with another data.

        :param other: The other GraphData object.
        :param return_dict: if true, return dictionary of which tensors share the same
            storage. Else returns true if any tensors share the same storage.
        :return:
        """
        d = {}
        for k, _ in self._tensors:
            a = getattr(self, k)
            b = getattr(other, k)
            if 0 in a.shape or 0 in b.shape:
                shares_storage = False
            else:
                shares_storage = same_storage(a, b)
            if return_dict:
                d[k] = shares_storage
            elif shares_storage:
                return True
        return d

    def contiguous(self):
        return self.apply(lambda x: x.contiguous())

    @property
    def num_graphs(self):
        return self.g.shape[0]

    @property
    def num_nodes(self):
        return self.x.shape[0]

    @property
    def num_edges(self):
        return self.edges.shape[1]

    @property
    def node_shape(self):
        return self.x.shape[1:]

    @property
    def edge_shape(self):
        return self.e.shape[1:]

    @property
    def global_shape(self):
        return self.g.shape[1:]

    @property
    def shape(self):
        return self.x.shape[1:] + self.e.shape[1:] + self.g.shape[1:]

    @property
    def size(self):
        return self.x.shape[:1] + self.e.shape[:1] + self.g.shape[:1]

    # def _mask_fields(self, masks: Dict[str, torch.tensor]):
    #     for m in masks:
    #         if m not in self.__slots__:
    #             raise RuntimeError("{} is not a valid field".format(m))
    #     masked_fields = []
    #     for field in self.__slots__:
    #         if field not in masks or masks[field] is None:
    #             masked_fields.append(getattr(self, field))
    #         else:
    #             masked_fields.append(getattr(self, field)[:, masks[field]])
    #     return masked_fields
    #
    # def mask(self, node_mask, edge_mask, global_mask, invert: bool = False):
    #     d = {"x": node_mask, "e": edge_mask, "g": global_mask}
    #     if invert:
    #         d = {k: ~v for k, v in d.items()}
    #     return self.__class__(*self._mask_fields(d))

    @staticmethod
    def _apply_mask(
        arr: torch.Tensor,
        mask: torch.BoolTensor,
        detach: bool,
        as_view: bool,
        dim: int = 0,
    ):
        ret = arr
        if detach:
            ret = arr.detach()

        if mask is None:
            if not as_view:
                ret = arr.clone()
        else:
            if as_view:
                mask = torch.where(mask)
            if isinstance(mask, tuple) or dim == 0:
                ret = arr[mask]
            elif dim == 1:
                ret = arr[:, mask]
            else:
                raise ValueError("dim must be 0 or 1")
        return ret

    def _validate_masks(
        self,
        node_mask: Optional[torch.BoolTensor],
        edge_mask: Optional[torch.BoolTensor],
    ):
        if node_mask is not None and not node_mask.ndim == 1:
            raise ValueError("Node mask must be 1 dimensional")
        if edge_mask is not None and not edge_mask.ndim == 1:
            raise ValueError("Edge mask must be 1 dimensional")
        if node_mask is not None and not node_mask.dtype == torch.bool:
            raise ValueError(
                "Node mask must be tensor.BoolTensor, not " + str(node_mask.dtype)
            )
        if edge_mask is not None and not edge_mask.dtype == torch.bool:
            raise ValueError(
                "Edge mask must be tensor.BoolTensor, not " + str(edge_mask.dtype)
            )

    def _mask_dispatch_reindex_edges(self, edges, node_mask):
        # remap node indices in edges
        if node_mask is not None and not torch.all(node_mask):
            nidx = torch.where(node_mask)[0]
            _, edges = reindex_tensor(nidx, edges)
        return edges

    def _mask_dispatch_constructor(self, new_inst: bool, *args, **kwargs):
        if new_inst:
            constructor = self.__class__
        else:

            def constructor(*args, **kwargs):
                self.__init__(*args, **kwargs)
                return self

        masked_data = constructor(*args, **kwargs)
        masked_data.debug()
        return masked_data

    def _mask_dispatch(
        self,
        node_mask: Optional[torch.BoolTensor],
        edge_mask: Optional[torch.BoolTensor],
        as_view: bool,
        detach: bool,
        new_inst: bool,
    ):
        self._validate_masks(node_mask, edge_mask)
        edges = self._apply_mask(self.edges, edge_mask, detach, as_view, dim=1)
        edges = self._mask_dispatch_reindex_edges(edges, node_mask)
        x = self._apply_mask(self.x, node_mask, detach, as_view)
        e = self._apply_mask(self.e, edge_mask, detach, as_view)
        g = self._apply_mask(self.g, None, detach, as_view)
        return self._mask_dispatch_constructor(new_inst, x, e, g, edges)

    def apply_edge_mask_(self, mask: torch.BoolTensor) -> GraphData:
        """In place version of :meth:`caldera.data.GraphData.apply_edge_mask`

        :param mask: boolean mask
        :return: self
        """
        return self._mask_dispatch(
            None, mask, as_view=True, detach=False, new_inst=False
        )

    def apply_edge_mask(self, mask: torch.BoolTensor) -> GraphData:
        """Apply edge mask to the graph, returning a new :class:`GraphData`
        instance.

        :param mask: boolean mask
        :return: a new :class:`GraphData` instance.
        """
        return self._mask_dispatch(
            None, mask, as_view=False, detach=True, new_inst=True
        )

    def _get_edge_mask_from_nodes(self, nodes: torch.LongTensor):
        idx = self._gather(self._node_to_edge_idx, nodes)
        mask = torch.BoolTensor([True] * self.num_edges)
        mask[idx] = False
        return mask

    def _apply_node_mask_dispatch(
        self, node_mask, as_view: bool, detach: bool, new_inst: bool
    ):
        nidx = torch.where(~node_mask)[0]
        edge_mask = ~torch.any(
            long_isin(self.edges.flatten(), nidx, invert=False).view(2, -1), 0
        )
        return self._mask_dispatch(
            node_mask, edge_mask, as_view=as_view, detach=detach, new_inst=new_inst
        )

    def apply_node_mask_(self, node_mask: torch.BoolTensor) -> GraphData:
        """In place version of :meth:`caldera.data.GraphData.apply_node_mask`.

        :param node_mask: boolean mask
        :return: self
        """
        return self._apply_node_mask_dispatch(
            node_mask, as_view=True, detach=False, new_inst=False
        )

    def apply_node_mask(self, node_mask: torch.BoolTensor) -> GraphData:
        """Apply node mask to the graph, returning a new :class:`GraphData`
        instance, removing any edges if necessary. Note this will reindex any
        indices in the data (e.g. `edges`)

        :param mask: boolean mask
        :return: a new :class:`GraphData` instance.
        """
        return self._apply_node_mask_dispatch(
            node_mask, as_view=False, detach=True, new_inst=True
        )

    @property
    def requires_grad(self):
        g = [getattr(self, k).requires_grad for k in self._differentiable]
        return all(g)

    @requires_grad.setter
    def requires_grad(self, v):
        def set_requires_grad(x):
            x.requires_grad = v

        self.apply_(set_requires_grad, keys=self._differentiable)

    # TODO: clone tests
    def clone(self):
        """Clones the data.

        Note that like the `clone()` method, this function will be
        recorded in the computation graph.
        """
        return self.apply(lambda x: x.clone())

    # TODO: copy tests
    def copy(self, non_blocking: bool = False, *emtpy_like_args, **emtpy_like_kwargs):

        """non_blocking (bool) – if True and this copy is between CPU and GPU,
        the copy may occur asynchronously with respect to the host. For other
        cases, this argument has no effect.

        :return:
        """
        return self.apply(
            lambda x: torch.empty_like(x, *emtpy_like_args, **emtpy_like_kwargs).copy_(
                x, non_blocking=non_blocking
            )
        )

    # TODO: nbexamples
    # TODO: handle undirected and hypergraphs
    # TODO: check that features are NUMPY rather than TORCH

    @staticmethod
    def _get_nx_feature_size(g: DirectedGraph, key: str, global_key: str = None):
        def collect_feature_shapes(datalist):
            shapes = []
            for x in datalist:
                data = x[-1]
                if key not in data:
                    shape = None
                else:
                    shape = tuple(data[key].shape)
                shapes.append(shape)
            return shapes

        def get_feature_shape(arr, m):
            shapes = set(collect_feature_shapes(arr))
            if arr is ...:
                return 0
            if len(shapes) > 1:
                if None in shapes:
                    raise RuntimeError(
                        "{x} features shapes must be the same. Found {shapes}. "
                        "At least one {x} is missing a feature '{key}'.".format(
                            x=m, shapes=shapes, key=key
                        )
                    )
                else:
                    raise RuntimeError(
                        "{x} features for '{key}' must be the same. Found {shapes}.".format(
                            x=m, shapes=shapes, key=key
                        )
                    )
            elif len(shapes) == 0:
                return 0
            elif list(shapes)[0] is None:
                return 0
            return reduce(lambda a, b: a * b, list(shapes)[0])

        n_feat = get_feature_shape(g.nodes(data=True), "node")
        e_feat = get_feature_shape(g.edges(data=True), "edge")
        g_feat = get_feature_shape(
            g.globals(data=True, global_key=global_key), "global"
        )

        return n_feat, e_feat, g_feat

    # TODO: validate features are all same size
    # TODO: estimate sizes of features

    @classmethod
    def from_networkx(
        cls,
        g: DirectedGraph,
        n_node_feat: Optional[int] = None,
        n_edge_feat: Optional[int] = None,
        n_glob_feat: Optional[int] = None,
        *,
        feature_key: str = "features",
        global_attr_key: str = None,
        requires_grad: Optional[bool] = None,
        dtype: str = torch.float32,
    ):
        """Create a new :class:`GraphData` from a networkx graph.

        :param g:
        :param n_node_feat:
        :param n_edge_feat:
        :param n_glob_feat:
        :param feature_key: The key to look for data in node, edge, and global data.
        :param global_attr_key: Key to look for global data.
        :return:
        """
        if not isinstance(g, nx.Graph) or not nx_is_directed(g):
            raise TypeError(
                "Graph must be a directed graph instance, not a '{}'. Convert to directed graph first.".format(
                    g.__class__.__name__
                )
            )
        gdata = g.get_global(global_attr_key)
        _n_node_feat, _n_edge_feat, _n_glob_feat = cls._get_nx_feature_size(
            g, feature_key, global_attr_key
        )
        n_node_feat = n_node_feat or _n_node_feat
        n_edge_feat = n_edge_feat or _n_edge_feat
        n_glob_feat = n_glob_feat or _n_glob_feat
        n_node_size, n_edge_size, n_glob_size = (
            g.number_of_nodes(),
            g.number_of_edges(),
            1,
        )
        node_attr = np.empty((n_node_size, n_node_feat))
        edge_attr = np.empty((n_edge_size, n_edge_feat))
        glob_attr = np.empty((n_glob_size, n_glob_feat))

        nodes = sorted(list(g.nodes(data=True)))
        ndict = {}
        # TODO: support n dim tensors (but they get flattened anyways...)
        for i, (n, ndata) in enumerate(nodes):
            node_attr[i] = ndata.get(feature_key, np.array([])).flatten()
            ndict[n] = i

        # TODO: method to change dtype?
        edges = np.empty((2, n_edge_size), dtype=np.float)
        for i, (n1, n2, edata) in enumerate(g.edges(data=True)):
            edges[:, i] = np.array([ndict[n1], ndict[n2]])
            edge_attr[i] = edata.get(feature_key, np.array([])).flatten()

        if feature_key in gdata:
            glob_attr[0] = gdata.get(feature_key, np.array([])).flatten()

        if requires_grad is not None:
            tensor = functools.partial(
                torch.tensor, dtype=dtype, requires_grad=requires_grad
            )
        else:
            tensor = functools.partial(torch.tensor, dtype=dtype)

        data = GraphData(
            tensor(node_attr),
            tensor(edge_attr),
            tensor(glob_attr),
            torch.tensor(edges, dtype=torch.long),
        )
        return data

    def to_networkx(
        self,
        feature_key: str = "features",
        global_attr_key: str = None,
        graph_type: Type[DirectedGraph] = nx.OrderedMultiDiGraph,
    ) -> DirectedGraph:
        g = graph_type()
        for n, ndata in enumerate(self.x):
            g.add_node(n, **{feature_key: ndata})
        g.ordered_edges = []
        for i, e in enumerate(self.edges.T):
            n = g.add_edge(e[0].item(), e[1].item(), **{feature_key: self.e[i]})
            g.ordered_edges.append((e[0].item(), e[1].item(), n))
        g.set_global({feature_key: self.g.clone()}, global_attr_key)
        return g

    def __repr__(self):
        return "<{cls} size(n,e,g)={size} features(n,e,g)={shape}>".format(
            cls=self.__class__.__name__,
            size=self.x.shape[:1] + self.e.shape[:1] + self.g.shape[:1],
            shape=self.shape,
        )

    def _eq_helper(
        self,
        other: "GraphData",
        comparator: Callable[[torch.Tensor, torch.Tensor], bool],
    ) -> bool:
        if not torch.all(torch.eq(self.edges, other.edges)):
            return False
        for attr in ["x", "e", "g"]:
            a = getattr(self, attr)
            b = getattr(other, attr)
            if not comparator(a, b):
                return False
        return True

    def __eq__(self, other: "GraphData") -> bool:
        def is_eq(a, b):
            if a.shape != b.shape:
                return False
            return torch.all(torch.eq(a, b))

        return self._eq_helper(other, comparator=is_eq)

    # TODO: implement this for batch?
    def append_nodes(self, node_attr: torch.Tensor):
        assert isinstance(self, GraphData)
        if not node_attr.ndim == 2:
            raise RuntimeError("Node attributes must have 2 dimensions")

        self.x = torch.cat([self.x, node_attr])
        self.debug()

    def append_edges(self, edge_attr: torch.Tensor, edges: torch.Tensor):
        assert isinstance(self, GraphData)
        self.edges = torch.cat([self.edges, edges], dim=1)
        self.e = torch.cat([self.e, edge_attr])
        self.debug()

    def allclose(self, other: "GraphData", **kwargs) -> bool:
        def _allclose(a, b):
            return torch.allclose(a, b, **kwargs)

        return self._eq_helper(other, comparator=_allclose)

    @classmethod
    def random(
        cls,
        n_feat: int,
        e_feat: int,
        g_feat: int,
        requires_grad: Optional[bool] = None,
        min_nodes: int = None,
        max_nodes: int = None,
        min_edges: int = None,
        max_edges: int = None,
    ) -> GraphData:
        # initialize defaults
        if min_nodes is None and max_nodes is None:
            min_nodes = 1
            max_nodes = 20
        elif min_nodes is None and max_nodes is not None:
            min_nodes = min(1, max_nodes)
        elif min_nodes is not None and max_nodes is None:
            max_nodes = max(min_nodes, 10)

        if min_edges is None and max_edges is None:
            min_edges = 1
            max_edges = max(min_edges, int(0.5 * max_nodes))
        elif min_edges is None and max_edges is not None:
            min_edges = min(1, max_edges)
        elif min_edges is not None and max_edges is None:
            max_edges = max(min_edges, int(0.5 * max_nodes))

        n_nodes = torch.randint(min_nodes, max_nodes + 1, torch.Size([])).item()
        n_edges = torch.randint(min_edges, max_edges + 1, torch.Size([])).item()

        edges = torch.empty((2, 0), dtype=torch.long)
        if n_nodes:
            edges = torch.randint(0, n_nodes, torch.Size([2, n_edges]))
        else:
            n_edges = 0
        return cls(
            torch.randn(n_nodes, n_feat),
            torch.randn(n_edges, e_feat),
            torch.randn(1, g_feat),
            edges,
            requires_grad=requires_grad,
        )

    # TODO: view
    def view(
        self,
        x_slice: Optional[slice] = None,
        e_slice: Optional[slice] = None,
        g_slice: Optional[slice] = None,
    ) -> GraphData:
        if x_slice is None:
            x_slice = slice(None, None, None)
        if e_slice is None:
            e_slice = slice(None, None, None)
        if g_slice is None:
            g_slice = slice(None, None, None)
        return self.__class__(
            self.x[:, x_slice], self.e[:, e_slice], self.g[:, g_slice], self.edges
        )

    def index_nodes(self, idx: torch.LongTensor) -> GraphData:
        """Apply index to nodes."""
        cloned = self.detach().clone()
        cloned.index_nodes_(idx)
        return cloned

    def index_nodes_(self, idx: torch.LongTensor) -> None:
        """In place version of :meth:`caldera.data.GraphData.index_nodes`"""
        assert idx.ndim == 1
        assert idx.shape[0] == self.num_nodes
        _, edges = reindex_tensor(idx, self.edges)
        x = self.x[idx]
        self.edges = edges
        self.x = x

    def index_edges(self, idx: torch.LongTensor) -> GraphData:
        """Apply index to nodes."""
        cloned = self.detach().clone()
        cloned.index_edges_(idx)
        return cloned

    def index_edges_(self, idx: torch.LongTensor) -> None:
        """In place version of :meth:`caldera.data.GraphData.index_edges`"""
        assert idx.ndim == 1
        assert idx.shape[0] == self.num_edges
        self.edges = self.edges[:, idx]
        self.e = self.e[idx]

    def shuffle_nodes(self) -> GraphData:
        cloned = self.detach().clone()
        cloned.shuffle_nodes_()
        return cloned

    def shuffle_edges(self) -> GraphData:
        cloned = self.detach().clone()
        cloned.shuffle_edges_()
        return cloned

    def shuffle_nodes_(self) -> None:
        idx = torch.randperm(self.num_nodes)
        self.index_nodes_(idx)

    def shuffle_edges_(self) -> None:
        idx = torch.randperm(self.num_edges)
        self.index_edges_(idx)

    def shuffle_(self) -> None:
        self.shuffle_edges_()
        self.shuffle_nodes_()

    def shuffle(self) -> GraphData:
        cloned = self.detach().clone()
        cloned.shuffle_()
        return cloned

    def reverse(self) -> GraphData:
        cloned = self.detach().clone()
        cloned.reverse_()
        return cloned

    def reverse_(self):
        self.edges = self.edges.flip(1)

    def nelement(self) -> int:
        """Return total number of elements in the.

        :class:`caldera.data.GraphData` instance
        """
        x = 0
        for _, t in self._tensors:
            if hasattr(t, "nelement"):
                x += t.nelement()
        return x

    def memsize(self):
        """Return total number of bytes in the.

        :class:`caldera.data.GraphData` instance
        """
        x = 0
        for _, t in self._tensors:
            if hasattr(t, "nelement"):
                x += t.element_size() * t.nelement()
        return x

    def density(self):
        """Return density of the graph."""
        return self.num_edges / (self.num_nodes * (self.num_nodes - 1))

    def _get_edge_dict(self):
        src, dest = self.edges.tolist()
        edge_dict = {}
        for _src, _dest in zip(src, dest):
            edge_dict.setdefault(_src, list())
            edge_dict[_src].append(_dest)
        return edge_dict

    def info(self):
        msg = "{}(\n".format(self.__class__.__name__)
        msg += "  n_nodes: {}\n".format(self.num_nodes)
        msg += "  n_edges: {}\n".format(self.num_edges)
        msg += "  feat_shape: {}\n".format(tuple(self.shape))
        msg += "  size: {}\n".format(tuple(self.size))
        msg += "  nelements: {}\n".format(self.nelement())
        msg += "  bytes: {}\n".format(self.memsize())
        msg += ")"
        return msg



"""_add_self_loops.py.

Add self loops to graph.
"""
from typing import overload

from ._base import TransformBase
from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.data.utils import add_edges


class AddSelfLoops(TransformBase):
    def __init__(self, fill_value: float = 0.0):
        super().__init__()
        self.fill_value = fill_value

    @overload
    def transform_each(self, data: GraphData) -> GraphData:
        ...

    def transform(self, data: GraphBatch) -> GraphBatch:  # noqa: E811
        return add_edges(data, self.fill_value, kind="self")



from ._all_pairs_shortest_path import floyd_warshall
from ._path_utils import PathAccumulator
from ._path_utils import PathMax
from ._path_utils import PathMin
from ._path_utils import PathMul
from ._path_utils import PathNpProduct
from ._path_utils import PathNpSum
from ._path_utils import PathSum
from ._path_utils import PathSymbol
from ._shortest_path import multisource_dijkstras

__all__ = [
    "floyd_warshall",
    "PathAccumulator",
    "PathMax",
    "PathMin",
    "PathMul",
    "PathNpProduct",
    "PathNpSum",
    "PathSum",
    "PathSymbol",
    "multisource_dijkstras",
]



import functools

from ._nx_feature_transform import NetworkxTransformFeatureData
from caldera.utils import dict_join


class NetworkxSetDefaultFeature(NetworkxTransformFeatureData):
    def __init__(self, node_default=None, edge_default=None, global_default=None):
        self.node_default = node_default
        self.edge_default = edge_default
        self.global_default = global_default

        setdefault = functools.partial(dict_join, join_fn=lambda a, b: a, mode="union")

        if node_default is not None:
            node_transform = functools.partial(setdefault, b=node_default)
        else:
            node_transform = None

        if edge_default is not None:
            edge_transform = functools.partial(setdefault, b=edge_default)
        else:
            edge_transform = None

        if global_default is not None:
            global_transform = functools.partial(setdefault, b=global_default)
        else:
            global_transform = None

        super().__init__(
            node_transform=node_transform,
            edge_transform=edge_transform,
            global_transform=global_transform,
        )



"""Adds the following to nx.Graph.

`get_global()`
"""
import functools

import networkx as nx

from caldera.utils.nx._global_accessor import GlobalAccess
from caldera.utils.nx._global_accessor import GraphWithGlobal


def add_global_access_to_nx(key):
    nx.Graph.get_global_key = functools.partialmethod(
        GraphWithGlobal.get_global_key, default_key=key
    )
    nx.Graph.get_global = GraphWithGlobal.get_global
    nx.Graph.set_global = GraphWithGlobal.set_global
    nx.Graph.globals = GraphWithGlobal.globals
    setattr(nx.Graph, key, GlobalAccess())



"""Adds the following to nx.Graph.

`get_global()`
"""
import networkx as nx


_global_key = "_global_data"


class GlobalDict(dict):
    def __init__(self, data, obj):
        super().__init__(data)
        self._obj = obj
        self._obj_set = False

    def update(self, d):
        if not self._obj_set:
            setattr(self._obj, _global_key, self)
        super().update(d)

    def setdefault(self, k, v):
        if not self._obj_set:
            setattr(self._obj, _global_key, self)
        super().__setitem__(k, v)

    def __setitem__(self, k, v):
        if not self._obj_set:
            setattr(self._obj, _global_key, self)
        super().__setitem__(k, v)


class GlobalAccess:
    def __get__(self, obj, objtype):
        if not hasattr(obj, _global_key):
            return GlobalDict({}, obj)
        return getattr(obj, _global_key)

    def __set__(self, obj, val):
        setattr(obj, _global_key, val)


def get_global(self, global_key: str = None, default_key: str = None):
    """Get.

    :param self:
    :param global_key:
    :param default_key:
    :return:
    """
    if global_key is None:
        global_key = default_key
    return getattr(self, global_key)


class GraphWithGlobal(nx.Graph):
    def get_global_key(self, global_key: str = None, default_key: str = None):
        """Retrieve the global_key for the graph. If `global_key` is provided,
        just return `global_key`. Else, use the provided default.

        Default key is set on `caldera` import (see `caldera._setup`) to be the value found
        in `caldera.defaults.CalderaDefaults.nx_global_key`.

        :param global_key:
        :param default_key:
        :return:
        """
        if global_key is None:
            global_key = default_key
        return global_key

    def get_global(self, global_key: str = None):
        """Get global attribute for the graph. If global_key not provided, will
        use the provided default_key.

        Default key is set on `caldera` import (see `caldera._setup`) to be the value found
        in `caldera.defaults.CalderaDefaults.nx_global_key`.

        :param global_key:
        :param default_key:
        :return:
        """
        gk = self.get_global_key(global_key)
        try:
            return getattr(self, gk)
        except AttributeError as e:
            msg = "'{}' object is missing attribute '{}' during call of `get_global`".format(
                self.__class__.__name__, gk
            )
            if global_key is None:
                msg += ". Global key was not explicitly provided."
            raise AttributeError(msg) from e

    def set_global(self, value: dict, global_key: str = None):
        """Set global attribute for the graph. If global_key not provided, will
        use the provided default_key.

        Default key is set on `caldera` import (see `caldera._setup`) to be the value found
        in `caldera.defaults.CalderaDefaults.nx_global_key`.

        :param global_key:
        :param default_key:
        :return:
        """
        if not isinstance(value, dict):
            raise TypeError("Value must be a `dict`")
        return setattr(self, self.get_global_key(global_key), value)

    def globals(self, data: bool = False, global_key: str = None):
        if data is True:
            yield self.get_global_key(global_key), self.get_global(global_key)
        else:
            yield self.get_global(global_key)



"""_base.py.

Transform base class
"""
from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import overload
from typing import Union

from caldera.data import GraphBatch
from caldera.data import GraphData


class TransformBase(ABC):
    @abstractmethod
    def transform(self, data):
        pass

    @overload
    def __call__(self, data: GraphBatch) -> GraphBatch:
        ...

    def __call__(self, data: GraphData) -> GraphData:
        return self.transform(data)

    def __repr__(self):
        return "{}".format(self.__class__.__name__)


TransformCallable = Union[Callable, TransformBase]



from typing import Any
from typing import Optional
from typing import Tuple

import torch
from torch import nn

from caldera.blocks.block import Block
from caldera.data import GraphData


class EdgeBlock(Block):
    def __init__(self, mlp: nn.Module):
        super().__init__({"mlp": mlp}, independent=True)

    def forward(self, edge_attr: torch.FloatTensor):
        results = self.block_dict["mlp"](edge_attr)
        return results

    def forward_from_data(self, data: GraphData):
        return self(data.e)


class AggregatingEdgeBlock(EdgeBlock):
    def __init__(self, mlp: nn.Module):
        super().__init__(mlp)
        self._independent = False

    def forward(
        self,
        *args: Tuple,
        edge_attr: torch.FloatTensor,
        node_attr: torch.FloatTensor,
        edges: torch.LongTensor,
        global_attr: Optional[torch.FloatTensor] = None,
        edge_idx: Optional[torch.LongTensor] = None,
    ) -> Any:
        to_agg = (node_attr[edges[0]], node_attr[edges[1]])
        if global_attr is not None:
            if edge_idx is None:
                raise RuntimeError(
                    "If `global_attr` provided must also provide `edge_index`"
                )
            to_agg += (global_attr[edge_idx],)
        out = torch.cat([*to_agg, edge_attr], 1)

        return self.block_dict["mlp"](out)

    def forward_from_data(self, data: GraphData):
        return self(data.e, data.x.data.edges)



import torch


class GraphNetworkBase(torch.nn.Module):
    """Base class for GraphNetwork modules."""

    # def reset_parameters(self):
    #     for child in self.children():
    #         if hasattr(child, 'reset_parameters'):
    #             child.reset_parameters()



# flake8: noqa
import networkx as nx
import pandas as pd
import seaborn as sns
import torch
from pylab import plt

from .graph_plotter import GraphPlotter


def plot_example_graphs(graphs, rows=4, cols=4, figsize=(10, 10)):
    import matplotlib as mplt

    norm = mplt.colors.Normalize(0, 20)

    fig, axes = plt.subplots(rows, cols, squeeze=False, figsize=figsize)

    def flatten(x):
        arr = []
        for a in x:
            for b in a:
                arr.append(b)
        return arr

    axes = flatten(axes)
    for g, ax in zip(graphs, axes):
        if not g.number_of_nodes():
            continue
        ylabels = {n: ndata for n, ndata in g.nodes(data="y")}
        nodelist, nodedata = zip(*g.nodes(data=True))
        ylabels = {
            n: round(ndata["y"].flatten().item(), 2)
            for n, ndata in zip(nodelist, nodedata)
        }
        y = [round(ndata["y"].flatten().item(), 2) for ndata in nodedata]

        pos = GraphPlotter._topological_sort(g)
        nx.draw_networkx_labels(g, pos, labels=ylabels, ax=ax)
        nx.draw_networkx_nodes(g, pos, ax=ax, node_color=mplt.cm.coolwarm(norm(y)))
        nx.draw_networkx_edges(g, pos, ax=ax)
    return fig, axes


def plot_kinetics(model, data, steps):
    with torch.no_grad():
        out = model(data, steps)
        x_arr = []
        for i, data in enumerate(out):
            t = torch.tensor([[i]] * data.x.shape[0], dtype=torch.float)
            index = torch.unsqueeze(torch.arange(data.x.shape[0], dtype=torch.float), 1)
            index = index.to(data.x.device)
            t = t.to(data.x.device)
            x = torch.cat([t, data.x, index], dim=1)
            x_arr.append(x)

        x = torch.cat(x_arr)
        df = pd.DataFrame(
            {
                "t": x[:, 0].detach().cpu(),
                "x": x[:, 1].detach().cpu(),
                "node": x[:, 2].detach().cpu(),
            }
        )
        df = df[df["node"] < 20]
        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca()
        ax.set_ylim(-5, 20)
        sns.lineplot(x="t", y="x", hue="node", data=df, ax=ax)

    return ax, fig



from typing import Callable
from typing import List

from torch import nn

from caldera.defaults import CalderaDefaults as D
from caldera.utils import pairwise


class MLPBlock(nn.Module):
    """A multilayer perceptron block."""

    def __init__(
        self,
        input_size: int,
        output_size: int = None,
        layer_norm: bool = True,
        dropout: float = None,
        activation: Callable = D.activation,
    ):
        """Initialize a multi-layer perceptron.

        :param input_size:
        :param output_size:
        :param layer_norm:
        :param dropout:
        :param activation:
        """
        super().__init__()
        if output_size is None:
            output_size = input_size
        layers = [nn.Linear(input_size, output_size), activation()]
        if layer_norm:
            layers.append(nn.LayerNorm(output_size))
        if dropout:
            layers.append(nn.Dropout(dropout))
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)


class MLP(nn.Module):
    """A multilayer perceptron."""

    def __init__(
        self,
        *latent_sizes: List[int],
        layer_norm: bool = True,
        dropout: float = None,
        activation: Callable = D.activation
    ):
        """

        :param latent_sizes:
        :param layer_norm:
        :param dropout:
        :param activation:
        """
        super().__init__()
        self.layers = nn.Sequential(
            *[
                MLPBlock(
                    n1,
                    n2,
                    layer_norm=layer_norm,
                    dropout=dropout,
                    activation=activation,
                )
                for n1, n2 in pairwise(latent_sizes)
            ]
        )

    def forward(self, x):
        return self.layers(x)



from typing import Union

import torch
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph._shortest_path import (
    floyd_warshall as cs_graph_floyd_warshall,
)

from ._sparse import to_sparse_coo_matrix
from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.utils.sparse import torch_coo_to_scipy_coo


def floyd_warshall(data: Union[GraphData, GraphBatch], **kwargs) -> torch.Tensor:
    """Run the floyd-warshall algorithm."""
    m = to_sparse_coo_matrix(data, fill_value=1).coalesce()
    graph = csr_matrix(torch_coo_to_scipy_coo(m))
    default_kwargs = dict(directed=True, unweighted=True)
    default_kwargs.update(kwargs)
    matrix = cs_graph_floyd_warshall(graph, **default_kwargs)
    return torch.from_numpy(matrix)



from caldera.blocks import EdgeBlock
from caldera.blocks import GlobalBlock
from caldera.blocks import NodeBlock
from caldera.data import GraphBatch
from caldera.data import GraphTuple
from caldera.models.base import GraphNetworkBase


class GraphEncoder(GraphNetworkBase):
    def __init__(
        self, edge_block: EdgeBlock, node_block: NodeBlock, global_block: GlobalBlock
    ):
        assert issubclass(type(edge_block), EdgeBlock)
        assert issubclass(type(node_block), NodeBlock)
        assert issubclass(type(global_block), GlobalBlock)
        super().__init__()
        self.node_block = node_block
        self.edge_block = edge_block
        self.global_block = global_block

    def forward(self, data: GraphBatch) -> GraphTuple:
        def run_block(block):
            try:
                return block.forward_from_data(data)
            except RuntimeError as e:
                raise type(e)(
                    "error running `{}.forward()`. {}".format(block._get_name(), str(e))
                ) from e

        edge_attr = run_block(self.edge_block)
        node_attr = run_block(self.node_block)
        global_attr = run_block(self.global_block)
        return GraphTuple(edge_attr, node_attr, global_attr)



import pytorch_lightning as pl
import torch
import wandb
from pytorch_lightning import LightningModule

from .configuration import Config
from .configuration.tools import dataclass_to_dict
from .model import Network
from .plotting import comparison_plot
from .plotting import fig_to_pil
from .plotting import figs_to_pils
from caldera.data import GraphBatch


class TrainingModule(LightningModule):
    def __init__(self, config: Config):
        super().__init__()
        self.config = config
        self.model: Network = Network(config.network)
        self.hparams = dataclass_to_dict(config)
        # self._loss_fn = torch.nn.BCEWithLogitsLoss(reduction='mean')

    def on_fit_start(self):
        self.logger.experiment.config.update(self.hparams)

    def training_step(self, batch: GraphBatch, batch_idx: int) -> pl.TrainResult:
        input_batch, target_batch = batch
        out_batch_list = self.model.forward(
            input_batch,
            steps=self.config.hyperparameters.train_core_processing_steps,
            save_all=True,
        )

        _loss_f = torch.nn.BCELoss()
        for out_batch in out_batch_list:
            node_loss = _loss_f(out_batch.x, target_batch.x)
            edge_loss = _loss_f(out_batch.e, target_batch.e)
            glob_loss = _loss_f(out_batch.g, target_batch.g)
        loss = (node_loss + edge_loss + glob_loss) / len(out_batch_list)

        result = pl.TrainResult(loss)
        result.log(
            "train_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=True
        )
        self.logger.experiment.log({"train_loss": loss})
        return pl.TrainResult(loss)

    def validation_step(self, batch, batch_idx):

        input_batch, target_batch = batch
        out_batch_list = self.model.forward(
            input_batch, steps=self.config.hyperparameters.train_core_processing_steps
        )
        _loss_f = torch.nn.BCELoss()
        for out_batch in out_batch_list:
            node_loss = _loss_f(out_batch.x, target_batch.x)
            edge_loss = _loss_f(out_batch.e, target_batch.e)
            glob_loss = _loss_f(out_batch.g, target_batch.g)
        loss = (node_loss + edge_loss + glob_loss) / len(out_batch_list)

        result = pl.EvalResult(checkpoint_on=loss)
        result.log(
            "eval_loss", loss, on_step=False, on_epoch=True, prog_bar=False, logger=True
        )

        if batch_idx == 0:
            self.validate_plot(batch)
        return result

    def validate_plot(self, batch, num_graphs=10):
        x, y = batch
        x.to_data_list()
        y.to_data_list()
        x = GraphBatch.from_data_list(x.to_data_list()[:num_graphs])
        y = GraphBatch.from_data_list(y.to_data_list()[:num_graphs])

        y_hat = self.model.forward(x, 10)[-1]
        y_graphs = y.to_networkx_list()
        y_hat_graphs = y_hat.to_networkx_list()

        figs = []
        for idx in range(len(y_graphs)):
            yg = y_graphs[idx]
            yhg = y_hat_graphs[idx]
            fig, axes = comparison_plot(yhg, yg)
            figs.append(fig)
        with figs_to_pils(figs) as pils:
            self.logger.experiment.log({"image": [wandb.Image(pil) for pil in pils]})

    # def validation_step(self, batch: Tuple[GraphBatch, GraphBatch], batch_idx: int) -> pl.EvalResult:
    #     x, y = batch
    #     y_hat = self.model(x)
    #     loss = F.cross_entropy(y_hat, y)
    #     acc = FM.accuracy(y_hat, y)
    #     result = pl.EvalResult(checkpoint_on=loss)
    #     result.log_dict({'val_acc': acc, 'val_loss': loss})
    #     return result

    # def test_step(self, batch, batch_idx) -> pl.EvalResult:
    #     result = self.validation_step(batch, batch_idx)
    #     result.rename_keys({'val_acc': 'test_acc', 'val_loss': 'test_loss'})
    #     return result

    def configure_optimizers(self):
        return torch.optim.Adam(
            self.model.parameters(), lr=self.config.hyperparameters.lr
        )



import functools
import inspect
import sys
from functools import wraps
from typing import Any
from typing import Dict
from typing import Generator
from typing import Generic
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar

import torch
from colorama import Fore

from caldera.exceptions import CalderaException

T = TypeVar("T")
M = TypeVar("M", bound=torch.nn.Module)


class InvalidFlexZeroDimension(Exception):
    """Flex dimension cannot be 0."""


class InvalidFlexNegativeDimension(Exception):
    """Flex dimension cannot be negative."""


class ResolveError(Exception):
    """There was an error resolving the module."""


class FlexDim:
    def __init__(self, pos: int = 0, dim: int = 1):
        """Flexible dimension to be used in conjunction with `FlexBlock`

        :param pos: position of the input arguments that contains the input data
        :param dim: dimension to use for the input shape
        """
        self.pos = pos
        self.dim = dim

    # TODO: resolve input_kwargs for FlexDim?
    def resolve(self, input_args, input_kwargs):
        d = input_args[self.pos].shape[self.dim]
        if d == 0:
            raise InvalidFlexZeroDimension("Dimension cannot be zero.")
        if d < 0:
            raise InvalidFlexNegativeDimension("Dimension cannot be less than zero.")
        return d

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self.pos, self.dim)


class FlexBlock(torch.nn.Module, Generic[M]):
    """Flexible Block that is resolve upon calling of `forward` with an
    example."""

    def __init__(self, module_fn: Type[M], *args, _from_frame=None, **kwargs):
        """A Flexible torch.nn.Module whose dimensions are resolved when
        provided with an example.

        :param module_fn:
        :param args:
        :param kwargs:
        """
        super().__init__()
        self.module: Type[M] = module_fn
        self.args = args
        self.kwargs = kwargs
        self.resolved_module: Optional[M] = None
        self._apply_history = None
        self.__resolved = False
        self._from_frame = _from_frame or str(sys._getframe(1))

        for fname in ["__call__", "forward"]:
            if hasattr(self.module, fname):
                setattr(
                    self,
                    fname,
                    functools.wraps(getattr(self.module, fname))(
                        functools.partial(getattr(self.__class__, fname), self)
                    ),
                )
                docstr = (
                    "Flex."
                    + fname
                    + " bound to function of "
                    + str(getattr(self.module, fname))
                )
                if getattr(self.module, fname).__doc__:
                    docstr += "\n" + getattr(self.module, fname).__doc__
                getattr(self, fname).__doc__ = docstr

    @property
    def is_resolved(self):
        """Returns whether this block has been resolved with an example. Note
        that before an example is presented, this block will have no
        parameters.

        :return: whether this block has been resolved
        """
        return self.__resolved

    def resolve_args(self, input_args: Tuple[Any, ...], input_kwargs: Dict[str, Any]):
        rargs = []
        for i, a in enumerate(self.args):
            if isinstance(a, FlexDim):
                rargs.append(a.resolve(input_args, input_kwargs))
            elif a is FlexDim:
                raise ValueError(
                    "Found {}. Initialize FlexDim to use flexible dimensions, `Flex.d()` or `FlexDim()`".format(
                        a
                    )
                )
            else:
                rargs.append(a)
        return rargs

    # TODO: implement resolve_kwargs
    def resolve_kwargs(self, input_args: Tuple[Any, ...], input_kwargs: Dict[str, Any]):
        return self.kwargs

    def resolve(self, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        resolved_args = self.resolve_args(args, kwargs)
        resolved_kwargs = self.resolve_kwargs(args, kwargs)
        self.__resolved = True
        try:
            self.resolved_module = self.module(*resolved_args, **resolved_kwargs)
        except Exception as e:
            msg = Fore.RED + "There was an error resolving module {}".format(
                self.module
            )
            msg += "\n  Flex or FlexBlock initialized in {}".format(self._from_frame)
            msg += "\nExpected Signature: {}".format(inspect.signature(self.module))
            tokens = [str(a) for a in resolved_args]
            tokens += ["{}={}".format(k, v) for k, v in resolved_kwargs.items()]
            msg += "\nReceived: ({})".format(", ".join(tokens))
            msg += "\nReceived the following exception:"
            msg += Fore.RESET
            msg += "\n"
            msg += str(e)
            raise ResolveError(msg) from e
        if self._apply_history:
            self._play_apply()

    def forward(self, *args, **kwargs):
        if not self.is_resolved:
            self.resolve(args, kwargs)
        return self.resolved_module(*args, **kwargs)

    def _record_apply(self, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        """Records the `_apply` function to the `_apply_history` storage."""
        if self._apply_history is None:
            self._apply_history = []
        self._apply_history.append((args, kwargs))

    def _play_apply(self):
        """Plays back the `_apply` function from the `_apply_history`
        storage."""
        try:
            for args, kwargs in self._apply_history:
                super()._apply(*args, **kwargs)
        except Exception as e:
            raise CalderaException(
                "An error occurred while trying to "
                " replay `_apply` in a {}. Try resolving the module"
                " by providing an example before using"
                " methods that use `_apply` (such as `to()`)".format(self.__class__)
            ) from e
        self._apply_history = None

    def _apply(self, *args, **kwargs) -> None:
        """Special override method. In the case that `_apply` is called and
        this module has not been resolved, the arguments for the `_apply`
        function are stored. These are reapplied after they have been resolved.

        :param args:
        :param kwargs:
        :return:
        """
        if not self.is_resolved:
            self._record_apply(args, kwargs)
        else:
            super()._apply(*args, **kwargs)

    # def _get_name(self):
    #     return "{}({})".format(
    #         self.__class__.__name__,
    #         self.module.__name__
    #     )

    def __repr__(self):
        if self.is_resolved:
            return super().__repr__()
        else:
            s = "{c}(\n\t(unresolved_module): {m}({args}, {kwargs}\n)".format(
                c=self._get_name(),
                m=self.module.__name__,
                args=", ".join([str(a) for a in self.args]),
                kwargs=",".join(str(k) + "=" + str(v) for k, v in self.kwargs.items()),
            )
            return s


def _iter_modules_of_type(
    module: torch.nn.Module, module_type: Type[T]
) -> Generator[T, None, None]:
    for m in module.modules():
        if issubclass(m.__class__, module_type):
            yield m


def _iter_flex_blocks(module: torch.nn.Module) -> Generator[FlexBlock, None, None]:
    yield from _iter_modules_of_type(module, FlexBlock)


# TODO: allow Flex to intake a function in addition to torch.nn.Module
class Flex(Generic[M]):
    """Flex."""

    d = FlexDim

    def __init__(self, module_type: Type[M]):
        """Initialize a module as a FlexBlock with flexible dimensions.

        Usage:

        .. code-block::

            Flex(torch.nn.Linear)(Flex.d(), 25)

        :param module_type: module type (e.g. `torch.nn.Linear`
        """
        self.module_type = module_type
        self._update_docstr()
        self._from_frame = str(sys._getframe(1))

    def _update_docstr(self):
        docstr = "Flex({m}). A module with flexible dimensions that wraps the {m} module.".format(
            m=self.module_type.__name__
        )
        docstr += "\nInitialize using Flex({m})(*args, **kwargs) according to the documentation below.\n".format(
            m=self.module_type.__name__
        )
        if self.module_type.__doc__:
            docstr += "\n{}\n".format(self.module_type.__name__)
            docstr += self.module_type.__doc__
        if self.module_type.__init__.__doc__:
            docstr += "\n__init__\n"
            docstr += self.module_type.__init__.__doc__
        self.__call__ = wraps(self.module_type)(self.__class__.__call__)
        self.__call__.__doc__ = docstr

    def __call__(self, *args, **kwargs) -> FlexBlock:
        """Initialize the flexible module.

        :param args: the initialization arguments
        :param kwargs: the initialization keyword arguments
        :return: initialized torch.nn.Module
        """
        return FlexBlock(
            self.module_type, *args, _from_frame=self._from_frame, **kwargs
        )

    @staticmethod
    def has_flex_blocks(module: torch.nn.Module):
        for m in _iter_flex_blocks(module):
            if issubclass(m.__class__, FlexBlock):
                return True

    @staticmethod
    def has_unresolved_flex_blocks(module: torch.nn.Module):
        for m in _iter_flex_blocks(module):
            if not m.is_resolved:
                return True
        return False



# __version__.py
# autogenerated by poetry-hooks 0.1.0

__version__ = "0.1.0a0"
__title__ = "caldera"
__authors__ = ["Justin Vrana <justin.vrana@gmail.com>"]
__repository__ = ""
__homepage__ = "http://www.github.com/jvrana/caldera"
__description__ = ""
__maintainers__ = ""
__readme__ = ""
__license__ = ""



from typing import Optional
from typing import Type

import torch
from scipy.sparse import coo_matrix

from .indexing import SizeType
from .indexing import unroll_index


def torch_coo_to_scipy_coo(m: torch.sparse.FloatTensor) -> coo_matrix:
    """Convert torch :class:`torch.sparse.FloatTensor` tensor to.

    :class:`scipy.sparse.coo_matrix`
    """
    data = m.values().numpy()
    indices = m.indices()
    return coo_matrix((data, (indices[0], indices[1])), tuple(m.size()))


def scatter_indices(indices: torch.LongTensor, shape: SizeType):
    """Unroll the coo indices using the provided shape.

    .. code-block::

        indices = torch.tensor([
            [0, 1, 2],
            [2, 3, 4],
            [4, 5, 4]
        ])
        shape = (3, 2)
        print(scatter_indices(indices, shape))

        # tensor([[0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2,
        #  0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2],
        # [2, 3, 4, 2, 3, 4, 2, 3, 4, 2, 3, 4, 2, 3, 4, 2, 3, 4, 2, 3, 4, 2, 3, 4,
        #  2, 3, 4, 2, 3, 4, 2, 3, 4, 2, 3, 4],
        # [4, 5, 4, 4, 5, 4, 4, 5, 4, 4, 5, 4, 4, 5, 4, 4, 5, 4, 4, 5, 4, 4, 5, 4,
        #  4, 5, 4, 4, 5, 4, 4, 5, 4, 4, 5, 4],
        # [0, 0, 1, 1, 2, 2, 0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 2, 2, 0, 1, 0, 1, 0, 1,
        #  0, 0, 1, 1, 2, 2, 0, 1, 0, 1, 0, 1]])

    :param indices:
    :param shape:
    :return:
    """
    if not shape:
        return indices
    idx = torch.stack(unroll_index(shape))

    a_repeat = [1] * indices.ndim
    a_repeat[-1] = idx.shape[-1]
    b_repeat = [1] * indices.ndim
    b_repeat[-1] = indices.shape[-1]

    a = torch.repeat_interleave(indices, idx.shape[-1], dim=1)
    b = idx.repeat(b_repeat)
    return torch.cat((a, b))


def _expand_idx(idx):
    if idx.ndim == 1:
        idx = idx.unsqueeze(0)
        idx = torch.cat((torch.zeros_like(idx), idx))
    return idx


def _coo_tensor(
    indices: torch.LongTensor,
    source: torch.Tensor,
    size: Optional[SizeType] = None,
    dtype: Optional[Type] = None,
    **kwargs
):
    if size is not None:
        kwargs["size"] = size
    if dtype is None:
        kwargs["dtype"] = source.dtype
    else:
        kwargs["dtype"] = dtype
    if size is not None:
        kwargs = dict(dtype=dtype, size=size)
    else:
        kwargs = dict(dtype=dtype)
    return torch.sparse_coo_tensor(indices, source, **kwargs)


# TODO: infer size from index sizes
def scatter_coo(
    indices: torch.LongTensor,
    source: torch.FloatTensor,
    size: Optional[SizeType] = None,
    expand: bool = False,
    dtype: Optional[Type] = None,
) -> torch.sparse.FloatTensor:
    """Scatter the provided source tensor to the provided indices.

    :param indices:
    :param source:
    :return:
    """

    indices = _expand_idx(indices)

    if not torch.is_tensor(source):
        source = torch.tensor(source)

    if expand:
        shape = source.shape
        # r = prod(shape[:-1]) * indices.shape[1]
        r = indices.shape[1]
        flattened = source.view(-1).repeat(r)
    else:
        shape = source.shape[1:]
        flattened = source.view(-1)

    if size is not None and size[-1] is ...:
        if not len(size) - 1 == indices.shape[0]:
            raise ValueError(
                "Provided dims ({}) must match number of index dims ({})".format(
                    len(size) - 1, indices.shape[0]
                )
            )
        size = tuple(list(size)[:-1]) + shape

    sidx = scatter_indices(indices, shape)
    return _coo_tensor(sidx, flattened, size=size, dtype=dtype)


#
# def scatter_coo_fill(
#     indices: torch.LongTensor,
#     source: torch.FloatTensor,
#     size: Optional[SizeType] = None,
#     dtype: Optional[Type] = None,
# ) -> torch.sparse.FloatTensor:
#     """Fill sparse coo matrix with the provided tensor at the provided indices.
#
#     :param indices:
#     :param source:
#     :return:
#     """
#     indices = _expand_idx(indices)
#     source = torch.tensor(source)
#     sidx = scatter_indices(indices, source.shape)
#     if size is not None and size[-1] is ...:
#         size = tuple(list(size)[:-1])
#         if torch.is_tensor():
#             size += source.shape
#     return _coo_tensor(
#         sidx, source.view(-1).repeat(indices.shape[1]), size=size, dtype=dtype
#     )



"""_compose.py.

Transform that composes other transforms. Exactly the same as
TorchVision interface.
"""
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import Union

from ._base import TransformCallable

T = TypeVar("T")


class Compose:
    def __init__(
        self, transforms: Union[List[TransformCallable], Tuple[TransformCallable, ...]]
    ):
        """Composes several transforms together.

        :param transforms:
        """
        self.transforms = transforms

    def __call__(self, data: T) -> T:
        for t in self.transforms:
            data = t(data)
        return data

    def __repr__(self):
        sep = "\n   "
        format_string = self.__class__.__name__ + "("
        format_string += sep.join([str(t) for t in self.transforms])
        format_string += "\n)"
        return format_string



from typing import Union

import torch

from caldera.data import GraphBatch
from caldera.data import GraphData


def graph_matrix(
    g: GraphData,
    dtype=torch.float,
    include_edge_attr: bool = True,
    fill_value: Union[int, float, torch.Tensor] = 0,
    edge_value: Union[int, float, torch.Tensor] = 1,
):
    edges = g.edges
    if include_edge_attr:
        shape = (g.num_nodes, g.num_nodes, g.e.shape[1])
    else:
        shape = (g.num_nodes, g.num_nodes)
    M = torch.full(shape, fill_value=fill_value, dtype=dtype)
    if include_edge_attr:
        v = g.e
    else:
        v = edge_value
    M[edges.unbind()] = v
    return M


def _degree_matrix_from_edges(
    edges: torch.LongTensor, n_nodes: int, i: int
) -> torch.LongTensor:
    D = torch.zeros(n_nodes, dtype=torch.long)
    a, b = torch.unique(edges[i], return_counts=True)
    D[(a, a)] = b
    return D


def in_degree_matrix_from_edges(
    edges: torch.LongTensor, n_nodes: int
) -> torch.LongTensor:
    return _degree_matrix_from_edges(edges, n_nodes, 1)


def out_degree_matrix_from_edges(
    edges: torch.LongTensor, n_nodes: int
) -> torch.LongTensor:
    return _degree_matrix_from_edges(edges, n_nodes, 0)


def adj_matrix(data: Union[GraphData, GraphBatch]) -> torch.LongTensor:
    return adj_matrix_from_edges(data.edges, data.num_nodes)


def in_degree(data: Union[GraphData, GraphBatch]) -> torch.LongTensor:
    return in_degree_matrix_from_edges(data.edges, data.num_nodes)


def out_degree(data: Union[GraphData, GraphBatch]) -> torch.LongTensor:
    return out_degree_matrix_from_edges(data.edges, data.num_nodes)


def adj_matrix_from_edges(edges: torch.LongTensor, n_nodes: int) -> torch.LongTensor:
    A = torch.zeros((n_nodes, n_nodes), dtype=torch.long)
    for i in edges.T:
        A[i[0], i[1]] += 1
    return A



from functools import partial
from itertools import permutations
from typing import Hashable
from typing import List
from typing import Tuple

import networkx as nx
import numpy as np
import torch

from caldera.data import GraphBatch
from caldera.data import GraphDataLoader


def has_cycle(g: nx.DiGraph) -> bool:
    try:
        nx.find_cycle(g)
        return True
    except nx.NetworkXNoCycle:
        return False


def to_one_hot(arr: torch.Tensor, mx: int):
    oh = torch.zeros((arr.shape[0], mx))
    for i, a in enumerate(arr):
        oh[i, a] = 1.0
    return oh


def sigmoid(x, a, kd, n, offset, inv):
    return a - a / (1.0 + np.exp((-x + kd) * n * inv)) + offset


# class Augment:
#     @staticmethod
#     def add_all_edges(g):
#         nodes = list(g.nodes())
#         for n1, n2 in itertools.product(nodes, repeat=2):
#             if n2 not in g[n1] or g[n1][n2] is None or len(g[n1][n2]) == 0:
#                 g.add_edge(n1, n2, features=torch.tensor([0.]), target=torch.tensor([0.]))
#         return g

# TODO: record data generation parameters
# TODO: record actual parameters
# TODO: record actual data
class CircuitGenerator:

    functions = {"sigmoid": sigmoid}

    def __init__(self, n_parts: int):
        self.n_parts = n_parts
        self.func_name = "sigmoid"
        assert self.func_name in self.functions
        params, labels = self.random_params(self.n_parts)
        self.params = params
        self.param_labels = params

    #         self.func_params = {
    #             'A': {
    #                 'min': 19,
    #                 'max': 20,
    #                 'distribution': 'uniform'
    #             },
    #             'K': {
    #                 'min': 19,
    #                 'max': 20,
    #                 'distribution': 'uniform'
    #             },
    #             'n': {
    #                 'min': 1.9,
    #                 'max': 2.,
    #                 'distribution': 'uniform'
    #             },
    #             'o': {
    #                 'min': 0,
    #                 'max': 'A.max() / 10.'
    #             }
    #         }

    @property
    def func(self):
        return self.functions[self.func_name]

    def random_params(self, num):
        A = np.random.uniform(1, 20, size=(num))
        K = np.random.uniform(1, 20, size=(num))
        n = np.random.uniform(1, 2, size=(num))
        o = np.random.uniform(0, A.max() / 10.0, size=(num))

        A = np.expand_dims(A, 1)
        K = np.expand_dims(K, 1)
        n = np.expand_dims(n, 1)
        o = np.expand_dims(o, 1)

        # choose repressor (1) or activator (-1)
        i = np.random.choice([1], (num, 1))

        # [n_parts, n_params]
        params = np.hstack([A, K, n, o, i])
        labels = np.array(["A", "K", "n", "o", "i"])
        return params, labels

    def steady_state(self, g, acc="sum", node_to_part=lambda x: int(x)):
        acc_dict = {"sum": lambda x: np.sum(np.concatenate(x))}

        # in topological order, we evaluate the sigmoid function at each node
        for node in nx.topological_sort(g):

            idx = node_to_part(node)
            # gather all parents
            # accumulate outputs 'y' using the provided accumulation function
            parents = list(g.predecessors(node))
            if not parents:
                p = np.expand_dims(self.params[idx : idx + 1].T, 2)
                x = np.array([[0.0]])
            else:
                a = []
                for p in parents:
                    _x = g.nodes[p]["y"]
                    a.append(_x)
                if len(a) and len(a[0]):
                    x = acc_dict[acc](a)
                    x = x.reshape(_x.shape)
                else:
                    x = torch.tensor([], dtype=torch.float)
            y = self._partial_func(self.func_name, x, idx)
            g.nodes[node]["y"] = y

    def _partial_func(self, fname, x, node: Hashable):
        return self.functions[fname](
            x, *tuple(np.expand_dims(self.params[node : node + 1].T, 2))
        )

    def annotate_graph_with_features(self, g: nx.DiGraph, include_target: bool = True):
        # one-hot encode the graph nodes
        one_hot_encoded = to_one_hot(torch.arange(0, self.n_parts), self.n_parts)

        new_g = nx.DiGraph()
        for n, data in g.nodes(data=True):
            new_g.add_node(n, **data)
        for n1, n2, edata in g.edges(data=True):
            edata["features"] = np.array([0.0])
            if include_target:
                edata["target"] = np.array([1.0])
            new_g.add_edge(n1, n2, **edata)
        if include_target:
            self.steady_state(new_g, node_to_part=lambda x: x[-1])
        for n, ndata in new_g.nodes(data=True):
            # convert this to ONE HOT!
            ndata["features"] = one_hot_encoded[list(n)[-1]]
            if include_target:
                ndata["target"] = torch.tensor(
                    [ndata["y"].flatten()], dtype=torch.float
                )
        new_g.data = {"features": torch.tensor([0])}
        if include_target:
            new_g.data["target"] = torch.tensor([0])

        return new_g

    @staticmethod
    def graph_from_nodes(
        nodes: List[Tuple[Hashable, Hashable, Hashable]]
    ) -> nx.DiGraph:
        g = nx.DiGraph()
        for n1, n2 in permutations(nodes, r=2):
            if n1[-1] in (n2[0], n2[1]):
                g.add_edge(tuple(n1), tuple(n2))
        return g

    def random_circuit(self, part_range: Tuple[int, int]):
        num_nodes = np.random.randint(*part_range)
        nodes = np.random.randint(0, self.n_parts, size=(num_nodes, 3))
        g = self.graph_from_nodes(nodes)
        return g

    def iter_random_circuit(
        self,
        limit: int,
        part_range: Tuple[int, int],
        cycles: bool = True,
        annotate: bool = False,
    ) -> nx.DiGraph:
        new_circuit = partial(self.random_circuit, part_range=part_range)
        for i in range(limit):
            c = new_circuit()
            if cycles is True:
                pass
            else:
                while has_cycle(c):
                    c = new_circuit()
            if annotate:
                c = self.annotate_graph_with_features(c)
            yield c


def split(arr, x):
    assert x > 0 and x <= 1
    i = int(len(arr) * x)
    j = len(arr) - i
    return arr[:i], arr[j:]


# new data
# ability to generalize larger circuits
def generate_data(
    generator,
    train_size,
    train_part_range,
    devtest_size,
    devtest_part_range,
    train__train_dev_split=0.9,
    dev_test_split=0.5,
):
    # training distribution
    training_distribution = list(
        generator.iter_random_circuit(
            train_size, train_part_range, cycles=False, annotate=True
        )
    )
    train_graphs, train_dev_graphs = split(
        training_distribution, train__train_dev_split
    )

    # dev/test distribution
    test_dev_distribution = list(
        generator.iter_random_circuit(
            devtest_size, devtest_part_range, cycles=False, annotate=True
        )
    )
    dev_graphs, test_graphs = split(test_dev_distribution, dev_test_split)
    return {
        "train": train_graphs,
        "train/dev": train_dev_graphs,
        "dev": dev_graphs,
        "test": test_graphs,
    }


# augmenting edges results in loss of learning ability
# [Augment.add_all_edges(g) for g in tqdm(graphs)]


def create_loader(generator, graphs, batch_size, shuffle, pin_memory=False):
    train_batch = GraphBatch.from_networkx_list(
        graphs, n_edge_feat=1, n_node_feat=generator.n_parts, n_glob_feat=1
    )
    target_batch = GraphBatch.from_networkx_list(
        graphs, n_edge_feat=16, n_node_feat=1, n_glob_feat=1, feature_key="target"
    )
    train_list = train_batch.to_data_list()
    target_list = target_batch.to_data_list()
    if batch_size is None:
        batch_size = len(train_list)
    return GraphDataLoader(
        list(zip(train_list, target_list)),
        batch_size=batch_size,
        shuffle=shuffle,
        pin_memory=pin_memory,
    )



r"""
Defaults (:mod:`caldera.defaults`)
==================================

Caldera defaults.
"""
import torch


class CalderaDefaults:

    activation = torch.nn.LeakyReLU
    nx_global_key = "_global_attr"



from dataclasses import dataclass

from .tools import ConfigObj


@dataclass
class HyperParamConfig(ConfigObj):

    lr: float = 1e-3  #: the learning rate
    train_core_processing_steps: int = (
        10  #: number of core processing steps for training step
    )
    eval_core_processing_steps: int = (
        10  #: number of core processing steps for evaluation step
    )



import networkx as nx

from caldera.transforms.networkx import NetworkxSetDefaultFeature
from caldera.utils.nx.generators import random_node


def annotate_shortest_path(
    g: nx.Graph,
    annotate_nodes: bool = True,
    annotate_edges: bool = True,
    source_key: str = "source",
    target_key: str = "target",
    path_key: str = "shortest_path",
    source: str = None,
    target: str = None,
) -> nx.Graph:
    """

    :param g:
    :param annotate_nodes:
    :param annotate_edges:
    :param source_key:
    :param target_key:
    :param path_key:
    :param source:
    :param target:
    :return:
    """
    if not annotate_edges and not annotate_nodes:
        raise ValueError("Must annotate either nodes or edges (or both)")
    source = source or random_node(g)
    target = target or random_node(g)

    g.nodes[source][source_key] = True
    g.nodes[target][target_key] = True

    try:
        path = nx.shortest_path(g, source=source, target=target)
    except nx.NetworkXNoPath:
        path = []

    NetworkxSetDefaultFeature(
        node_default={target_key: False, source_key: False, path_key: False},
        edge_default={path_key: False},
    )

    if annotate_nodes:
        for n in path:
            g.nodes[n][path_key] = True

    if annotate_edges:
        for n1, n2 in nx.utils.pairwise(path):
            g.edges[(n1, n2)][path_key] = True


# def draw_shortest_path(g):
#     pprint(feature_info(g))
#     fig = plt.figure(figsize=(3, 3))
#     ax = fig.gca()
#     ax.axis('off')
#     nodelist = list(g.nodes)
#
#     node_color = []
#     for n in nodelist:
#         node_color.append(g.nodes[n]['shortest_path'])
#
#     edge_list = []
#     edge_color = []
#     for n1, n2, edata in g.edges(data=True):
#         edge_list.append((n1, n2))
#         edge_color.append(edata['shortest_path'])
#
#     pos = nx.layout.spring_layout(g)
#     nx.draw_networkx_nodes(g, pos=pos, node_color=node_color, node_size=10, ax=ax)
#     nx.draw_networkx_edges(g, pos=pos, width=0.5, edge_color=edge_color, ax=ax)



from ._nx_feature_transform import NetworkxTransformFeatures
from caldera.utils import functional


class NetworkxNodesToStr(NetworkxTransformFeatures):
    def __init__(self):
        """Convert node keys to strings."""
        super().__init__(
            node_transform=functional.map_each(lambda x: (str(x[0]), x[1]))
        )



# flake8: noqa
import os
import sys

here = os.path.abspath(os.path.dirname(__file__))

PACKAGE_PARENT = "../.."
SCRIPT_DIR = os.path.dirname(
    os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
)
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))

import wandb
import torch
from caldera.data import GraphBatch
from typing import Tuple
from typing import Dict
from data import generate_data, create_loader, CircuitGenerator
from model import Network
from summary import loader_summary
from tqdm.auto import tqdm
from caldera.utils import _first
import pylab as plt
import pandas as pd
import seaborn as sns
import numpy as np
import random

import argparse

defaults = dict(
    latent_size_0=512,
    latent_size_1=512,
    latent_size_2=512,
    latent_depth_0=5,
    latent_depth_1=5,
    latent_depth_2=5,
    processing_steps=5,
    pass_global_to_node=True,
    pass_global_to_edge=True,
    lr_to_batch_size=5e-4 / 3e3,
    batch_size=3000,
    weight_decay=1e-3,
    epochs=1000,
    train_size=10000,
    dev_size=2000,
    log_every_epoch=10,
    dropout=0.2,
)


def create_data(
    train_size: int,
    dev_size: int,
    n_parts: int,
    train_part_range: Tuple[int, int],
    dev_part_range: Tuple[int, int],
) -> Dict:
    circuit_gen = CircuitGenerator(n_parts)

    data = generate_data(
        circuit_gen, train_size, train_part_range, dev_size, dev_part_range
    )

    return data, circuit_gen


def to(batch, device, **kwargs):
    return GraphBatch(
        batch.x.to(device, **kwargs),
        batch.e.to(device, **kwargs),
        batch.g.to(device, **kwargs),
        batch.edges.to(device, **kwargs),
        batch.node_idx.to(device, **kwargs),
        batch.edge_idx.to(device, **kwargs),
    )


def plot(target_data, out):
    x = target_data.x.cpu().detach().numpy().flatten()
    y = out.x.cpu().detach().numpy().flatten()
    fig = plt.figure(figsize=(10, 10))
    ax = fig.gca()
    df = pd.DataFrame({"x": x, "y": y})
    ax = sns.scatterplot("x", "y", data=df, ax=ax)
    ax.set_ylim(-5, 20)
    ax.set_xlim(-5, 20)
    return ax, fig


def seed(s):
    random.seed(s)
    np.random.seed(s)
    torch.manual_seed(s)


def train(**kwargs):
    """This is the documentation for the function.

    :param learning_rate:
    :return:
    """

    # Set up your default hyperparameters before wandb.init
    # so they get properly set in the train
    hyperparameter_defaults = dict(defaults)

    # Pass your defaults to wandb.init

    hyperparameter_defaults.update(kwargs)
    wandb.init(config=hyperparameter_defaults)
    config = wandb.config

    # update config
    wandb.config.update(
        {"learning_rate": config["batch_size"] * config["lr_to_batch_size"]}
    )

    # Your model here ...
    device = "cuda:0"

    class Null:
        pass

    net_config = dict(
        latent_sizes=(
            config["latent_size_0"],
            config["latent_size_1"],
            config["latent_size_2"],
        ),
        depths=(
            config["latent_depth_0"],
            config["latent_depth_1"],
            config["latent_depth_2"],
        ),
        pass_global_to_edge=config["pass_global_to_edge"],
        pass_global_to_node=config["pass_global_to_node"],
        dropout=config["dropout"],
        edge_to_node_aggregators=config.get("aggregators", Null),
        edge_to_global_aggregators=config.get("aggregators", Null),
        node_to_global_aggregators=config.get("aggregators", Null),
        aggregator_activation=config.get("aggregator_activation", Null),
    )
    net_config = {k: v for k, v in net_config.items() if v is not Null}

    net_config["aggregator_activation"] = {
        "leakyrelu": torch.nn.LeakyReLU,
        "tanh": torch.nn.Tanh,
        "sigmoid": torch.nn.Sigmoid,
    }.get(net_config["aggregator_activation"])

    net = Network(**net_config)

    # create your data
    data, gen = create_data(
        config["train_size"], config["dev_size"], 20, (2, 6), (8, 20)
    )

    seed(0)
    train_loader = create_loader(
        gen, data["train"], config["batch_size"], shuffle=True, pin_memory=True
    )
    seed(1)
    eval_loader = create_loader(
        gen, data["train/dev"], None, shuffle=False, pin_memory=True
    )

    wandb.config.update(
        {
            "train": {"loader_summary": loader_summary(train_loader)},
            "eval": {"loader_summary": loader_summary(eval_loader)},
        }
    )

    with torch.no_grad():
        batch, _ = _first(train_loader)
        out = net(batch, 3)
    net.to(device, non_blocking=True)

    assert list(net.parameters())
    optimizer = torch.optim.AdamW(
        net.parameters(),
        lr=config["learning_rate"],
        weight_decay=config["weight_decay"],
    )
    loss_fn = torch.nn.MSELoss()
    log_every_epoch = config["log_every_epoch"]
    for epoch in tqdm(range(config["epochs"]), desc="epochs"):
        net.train()

        running_loss = torch.tensor(0.0, device=device)
        for batch_idx, (train_data, target_data) in enumerate(train_loader):

            optimizer.zero_grad()

            # train_data.contiguous()
            # target_data.contiguous()
            # TODO: why clone?
            #             target_data = target_data.clone()
            train_data = to(train_data, device, non_blocking=True)
            target_data = to(target_data, device, non_blocking=True)
            out = net(train_data, config["processing_steps"])
            assert out[-1].x.shape == target_data.x.shape

            # TODO: the scale of the loss is proporational to the processing steps and batch_size, should this be normalized???
            loss = torch.tensor(0.0).to(
                device=device, non_blocking=True, dtype=torch.float32
            )
            for _out in out:
                loss += loss_fn(_out.x, target_data.x)
            loss = loss / (target_data.x.shape[0] * 1.0 * len(out))

            loss.backward()

            optimizer.step()

            loss.detach_()
            running_loss = running_loss + loss

        if epoch % log_every_epoch == 0:
            wandb.log({"train_loss": running_loss.cpu()}, step=epoch)

        if epoch % log_every_epoch == 0:
            net.eval()
            with torch.no_grad():
                eval_data, eval_target = _first(eval_loader)
                eval_data.detach_()
                eval_target.detach_()
                eval_data = to(eval_data, device, non_blocking=True)
                eval_target = to(eval_target, device, non_blocking=True)
                eval_outs = net(eval_data, config["processing_steps"], save_all=False)
                eval_outs[-1].detach_()
                eval_loss = (
                    loss_fn(eval_outs[-1].x, eval_target.x) / eval_outs[-1].x.shape[0]
                )

                wandb.log({"eval_loss": eval_loss}, step=epoch)

                wandb.log(
                    {
                        "edge_attr": wandb.Histogram(eval_outs[-1].e.cpu()),
                        "node_attr": wandb.Histogram(eval_outs[-1].x.cpu()),
                    },
                    step=epoch,
                )

                # # plot values
                # ax, fig = plot(eval_target, eval_outs[-1])
                # wandb.log({"chart": fig}, step=epoch)


if __name__ == "__main__":
    train()



from typing import Callable
from typing import Dict
from typing import Hashable
from typing import List
from typing import Set
from typing import Tuple

import torch


def _edges_to_tuples_set(edges: torch.LongTensor) -> Set[Tuple[int, int]]:
    edge_tuples = set()
    for e in edges.T:
        edge_tuples.add((e[0].item(), e[1].item()))
    return edge_tuples


def _validate_edges(edges: torch.LongTensor):
    assert edges.ndim == 2
    assert edges.shape[0] == 2


def _tuples_set_to_tensor(tuples: List[Tuple[int, int]]):
    return torch.tensor(list(zip(*list(tuples))), dtype=torch.long)


def _apply_to_edge_sets(
    edges1: torch.LongTensor,
    edges2: torch.LongTensor,
    func: Callable[[Set[Tuple[int, int]], Set[Tuple[int, int]]], torch.LongTensor],
) -> torch.LongTensor:
    s1 = _edges_to_tuples_set(edges1)
    s2 = _edges_to_tuples_set(edges2)
    s3 = func(s1, s2)
    return _tuples_set_to_tensor(s3)


def edges_difference(e1: torch.LongTensor, e2: torch.LongTensor) -> torch.LongTensor:
    def difference(e1, e2):
        return e1.difference(e2)

    return _apply_to_edge_sets(e1, e2, difference)


def edges_intersection(e1: torch.LongTensor, e2: torch.LongTensor) -> torch.LongTensor:
    def intersection(e1, e2):
        return e1.intersection(e2)

    return _apply_to_edge_sets(e1, e2, intersection)


def _edge_difference(edges1, edges2):
    s1 = _edges_to_tuples_set(edges1)
    s2 = _edges_to_tuples_set(edges2)
    s3 = s1.difference(s2)
    return s3


# @overload
# def add_edges(data: GraphBatch, fill_value: ..., kind: ...) -> GraphBatch:
#     """
#     Adds edges to the :class:`caldera.data.GraphBatch` instance.
#     """
#     ...


def get_edge_dict(edges: torch.LongTensor) -> Dict[Hashable, Set[Hashable]]:
    src, dest = edges.tolist()
    edge_dict = {}
    for _src, _dest in zip(src, dest):
        edge_dict.setdefault(_src, set())
        edge_dict[_src].add(_dest)
    return edge_dict



# flake8: noqa
##########################################################
# Relative Imports
##########################################################
import sys
from os.path import isfile
from os.path import join


def find_pkg(name: str, depth: int):
    if depth <= 0:
        ret = None
    else:
        d = [".."] * depth
        path_parts = d + [name, "__init__.py"]

        if isfile(join(*path_parts)):
            ret = d
        else:
            ret = find_pkg(name, depth - 1)
    return ret


def find_and_ins_syspath(name: str, depth: int):
    path_parts = find_pkg(name, depth)
    if path_parts is None:
        raise RuntimeError("Could not find {}. Try increasing depth.".format(name))
    path = join(*path_parts)
    if path not in sys.path:
        sys.path.insert(0, path)


try:
    import caldera
except ImportError:
    find_and_ins_syspath("caldera", 3)

##########################################################
# Main
##########################################################

import copy
import hydra
from examples.traversals.training import TrainingModule
from examples.traversals.data import DataGenerator, DataConfig
from examples.traversals.configuration import Config
from examples.traversals.configuration.data import Uniform, DiscreteUniform
from typing import TypeVar
from pytorch_lightning import Trainer
from examples.traversals.loggers import logger
from omegaconf import DictConfig, OmegaConf
from rich.panel import Panel
from rich import print
from rich.syntax import Syntax


C = TypeVar("C")


def prime_the_model(model: TrainingModule, config: Config):
    logger.info("Priming the model with data")
    config_copy: DataConfig = copy.deepcopy(config.data)
    config_copy.train.num_graphs = 10
    config_copy.eval.num_graphs = 0
    data_copy = DataGenerator(config_copy, progress_bar=False)
    for a, b in data_copy.train_loader():
        model.model.forward(a, 10)
        break


def print_title():
    print(Panel("Training Example: [red]Traversal", title="[red]caldera", expand=False))


def print_model(model: TrainingModule):
    print(Panel("Network", expand=False))
    print(model)


def print_yaml(cfg: Config):
    print(Panel("Configuration", expand=False))
    print(Syntax(OmegaConf.to_yaml(cfg), "yaml"))


# def config_override(cfg: DictConfig):
#     # defaults
#     cfg.hyperparameters.lr = 1e-3
#     cfg.hyperparameters.train_core_processing_steps = 10
#     cfg.hyperparameters.eval_core_processing_steps = 10
#
#     cfg.data.train.num_graphs = 5000
#     cfg.data.train.num_nodes = DiscreteUniform(10, 100)
#     cfg.data.train.density = Uniform(0.01, 0.03)
#     cfg.data.train.path_length = DiscreteUniform(5, 10)
#     cfg.data.train.composition_density = Uniform(0.01, 0.02)
#     cfg.data.train.batch_size = 512
#     cfg.data.train.shuffle = False
#
#     cfg.data.eval.num_graphs = 500
#     cfg.data.eval.num_nodes = DiscreteUniform(10, 100)
#     cfg.data.eval.density = Uniform(0.01, 0.03)
#     cfg.data.eval.path_length = DiscreteUniform(5, 10)
#     cfg.data.eval.composition_density = Uniform(0.01, 0.02)
#     cfg.data.eval.batch_size = "${data.eval.num_graphs}"
#     cfg.data.eval.shuffle = False


@hydra.main(config_path="conf", config_name="config")
def main(hydra_cfg: DictConfig):

    print_title()

    logger.setLevel(hydra_cfg.log_level)
    if hydra_cfg.log_level.upper() == "DEBUG":
        verbose = True
    else:
        verbose = False
    # really unclear why hydra has so many unclear validation issues with structure configs using ConfigStore
    # this correctly assigns the correct structured config
    # and updates from the passed hydra config
    # annoying... but this resolves all these issues
    cfg = OmegaConf.structured(Config())
    cfg.update(hydra_cfg)

    # debug
    if verbose:
        print_yaml(cfg)

    from pytorch_lightning.loggers import WandbLogger

    # explicitly convert the DictConfig back to Config object
    # has the added benefit of performing validation upfront
    # before any expensive training or logging initiates
    config = Config.from_dict_config(cfg)

    wandb_logger = WandbLogger(project="pytorchlightning", offline=config.offline)

    # initialize the training module
    training_module = TrainingModule(config)

    logger.info("Priming the model with data")
    prime_the_model(training_module, config)
    logger.debug(Panel("Model", expand=False))

    if verbose:
        print_model(training_module)

    logger.info("Generating data...")
    data = DataGenerator(config.data)
    data.init()

    logger.info("Beginning training...")
    trainer = Trainer(gpus=config.gpus, logger=wandb_logger, check_val_every_n_epoch=5)

    trainer.fit(
        training_module,
        train_dataloader=data.train_loader(),
        val_dataloaders=data.eval_loader(),
    )


if __name__ == "__main__":
    main()



from caldera.blocks import AggregatingEdgeBlock
from caldera.blocks import AggregatingGlobalBlock
from caldera.blocks import AggregatingNodeBlock
from caldera.data import GraphBatch
from caldera.data import GraphTuple
from caldera.models.base import GraphNetworkBase


class GraphCore(GraphNetworkBase):
    def __init__(
        self,
        edge_block: AggregatingEdgeBlock,
        node_block: AggregatingNodeBlock,
        global_block: AggregatingGlobalBlock,
        pass_global_to_edge: bool = False,
        pass_global_to_node: bool = False,
    ):
        assert issubclass(type(edge_block), AggregatingEdgeBlock)
        assert issubclass(type(node_block), AggregatingNodeBlock)
        assert issubclass(type(global_block), AggregatingGlobalBlock)
        super().__init__()
        self.node_block = node_block
        self.edge_block = edge_block
        self.global_block = global_block
        self.pass_to_global_to_edge = pass_global_to_edge
        self.pass_to_global_to_node = pass_global_to_node

    def forward(self, data: GraphBatch) -> GraphTuple:
        if self.pass_to_global_to_edge:
            edge_attr = self.edge_block(
                edge_attr=data.e,
                node_attr=data.x,
                edges=data.edges,
                global_attr=data.g,
                edge_idx=data.edge_idx,
            )
        else:
            edge_attr = self.edge_block(
                edge_attr=data.e, node_attr=data.x, edges=data.edges
            )

        if self.pass_to_global_to_node:
            node_attr = self.node_block(
                node_attr=data.x,
                edge_attr=edge_attr,
                edges=data.edges,
                global_attr=data.g,
                node_idx=data.node_idx,
            )
        else:
            node_attr = self.node_block(
                node_attr=data.x, edge_attr=edge_attr, edges=data.edges
            )

        global_attr = self.global_block(
            global_attr=data.g,
            node_attr=node_attr,
            edge_attr=edge_attr,
            edges=data.edges,
            node_idx=data.node_idx,
            edge_idx=data.edge_idx,
        )
        return GraphTuple(edge_attr, node_attr, global_attr)



import torch

from caldera.blocks import AggregatingEdgeBlock
from caldera.blocks import AggregatingGlobalBlock
from caldera.blocks import AggregatingNodeBlock
from caldera.blocks import EdgeBlock
from caldera.blocks import Flex
from caldera.blocks import GlobalBlock
from caldera.blocks import MLP
from caldera.blocks import MultiAggregator
from caldera.blocks import NodeBlock
from caldera.data import GraphBatch
from caldera.defaults import CalderaDefaults as defaults
from caldera.models import GraphCore
from caldera.models import GraphEncoder


class Network(torch.nn.Module):
    def __init__(
        self,
        latent_sizes=(128, 128, 1),
        depths=(1, 1, 1),
        dropout: float = None,
        pass_global_to_edge: bool = True,
        pass_global_to_node: bool = True,
        edge_to_node_aggregators=tuple(["add", "max", "mean", "min"]),
        edge_to_global_aggregators=tuple(["add", "max", "mean", "min"]),
        node_to_global_aggregators=tuple(["add", "max", "mean", "min"]),
        aggregator_activation=defaults.activation,
    ):
        super().__init__()
        self.config = {
            "latent_size": {
                "node": latent_sizes[1],
                "edge": latent_sizes[0],
                "global": latent_sizes[2],
                "core_node_block_depth": depths[0],
                "core_edge_block_depth": depths[1],
                "core_global_block_depth": depths[2],
            },
            "node_block_aggregator": edge_to_node_aggregators,
            "global_block_to_node_aggregator": node_to_global_aggregators,
            "global_block_to_edge_aggregator": edge_to_global_aggregators,
            "aggregator_activation": aggregator_activation,
            "pass_global_to_edge": pass_global_to_edge,
            "pass_global_to_node": pass_global_to_node,
        }
        self.encoder = GraphEncoder(
            EdgeBlock(Flex(MLP)(Flex.d(), latent_sizes[0], dropout=dropout)),
            NodeBlock(Flex(MLP)(Flex.d(), latent_sizes[1], dropout=dropout)),
            GlobalBlock(Flex(MLP)(Flex.d(), latent_sizes[2], dropout=dropout)),
        )

        edge_layers = [self.config["latent_size"]["edge"]] * self.config["latent_size"][
            "core_edge_block_depth"
        ]
        node_layers = [self.config["latent_size"]["node"]] * self.config["latent_size"][
            "core_node_block_depth"
        ]
        global_layers = [self.config["latent_size"]["global"]] * self.config[
            "latent_size"
        ]["core_global_block_depth"]

        self.core = GraphCore(
            AggregatingEdgeBlock(
                torch.nn.Sequential(
                    Flex(MLP)(Flex.d(), *edge_layers, dropout=dropout, layer_norm=True),
                    # Flex(torch.nn.Linear)(Flex.d(), edge_layers[-1])
                )
            ),
            AggregatingNodeBlock(
                torch.nn.Sequential(
                    Flex(MLP)(Flex.d(), *node_layers, dropout=dropout, layer_norm=True),
                    # Flex(torch.nn.Linear)(Flex.d(), node_layers[-1])
                ),
                Flex(MultiAggregator)(
                    Flex.d(),
                    self.config["node_block_aggregator"],
                    activation=self.config["aggregator_activation"],
                ),
            ),
            AggregatingGlobalBlock(
                torch.nn.Sequential(
                    Flex(MLP)(
                        Flex.d(), *global_layers, dropout=dropout, layer_norm=True
                    ),
                    # Flex(torch.nn.Linear)(Flex.d(), global_layers[-1])
                ),
                edge_aggregator=Flex(MultiAggregator)(
                    Flex.d(),
                    self.config["global_block_to_edge_aggregator"],
                    activation=self.config["aggregator_activation"],
                ),
                node_aggregator=Flex(MultiAggregator)(
                    Flex.d(),
                    self.config["global_block_to_node_aggregator"],
                    activation=self.config["aggregator_activation"],
                ),
            ),
            pass_global_to_edge=self.config["pass_global_to_edge"],
            pass_global_to_node=self.config["pass_global_to_node"],
        )

        self.decoder = GraphEncoder(
            EdgeBlock(
                Flex(MLP)(Flex.d(), latent_sizes[0], latent_sizes[0], dropout=dropout)
            ),
            NodeBlock(
                Flex(MLP)(Flex.d(), latent_sizes[1], latent_sizes[1], dropout=dropout)
            ),
            GlobalBlock(Flex(MLP)(Flex.d(), latent_sizes[2])),
        )

        self.output_transform = GraphEncoder(
            EdgeBlock(Flex(torch.nn.Linear)(Flex.d(), 1)),
            NodeBlock(Flex(torch.nn.Linear)(Flex.d(), 1)),
            GlobalBlock(Flex(torch.nn.Linear)(Flex.d(), 1)),
        )

    def forward(self, data, steps, save_all: bool = False):
        # encoded
        e, x, g = self.encoder(data)
        data = GraphBatch(x, e, g, data.edges, data.node_idx, data.edge_idx)

        # graph topography data
        edges = data.edges
        node_idx = data.node_idx
        edge_idx = data.edge_idx
        latent0 = data

        meta = (edges, node_idx, edge_idx)

        outputs = []
        for _ in range(steps):
            # core processing step
            e = torch.cat([latent0.e, e], dim=1)
            x = torch.cat([latent0.x, x], dim=1)
            g = torch.cat([latent0.g, g], dim=1)
            data = GraphBatch(x, e, g, *meta)
            e, x, g = self.core(data)

            # decode
            data = GraphBatch(x, e, g, *meta)

            _e, _x, _g = self.decoder(data)
            decoded = GraphBatch(_x, _e, _g, *meta)

            # transform
            _e, _x, _g = self.output_transform(decoded)
            print()
            gt = GraphBatch(_x, _e, _g, edges, node_idx, edge_idx)
            if save_all:
                outputs.append(gt)
            else:
                outputs = [gt]

        return outputs



from abc import ABC
from typing import Dict

from torch import nn


class Block(nn.Module, ABC):
    def __init__(self, module_dict: Dict[str, nn.Module], independent: bool):
        super().__init__()
        self._independent = independent
        self.block_dict = nn.ModuleDict(
            {name: mod for name, mod in module_dict.items() if mod is not None}
        )

    @property
    def independent(self):
        return self._independent

    # def reset_parameters(self):
    #     for child in self.children():
    #         if hasattr(child, 'reset_parameters'):
    #             child.reset_parameters()



import torch

from caldera.blocks.block import Block
from caldera.data import GraphBatch


class GlobalBlock(Block):
    def __init__(self, mlp):
        super().__init__({"mlp": mlp}, independent=True)

    def forward(self, global_attr):
        return self.block_dict["mlp"](global_attr)

    def forward_from_data(self, data: GraphBatch):
        return self(data.g)


# TODO: determine which aggregator to use during training (some function of attributes -> one-hot)
class AggregatingGlobalBlock(GlobalBlock):
    def __init__(self, mlp, edge_aggregator=None, node_aggregator=None):
        super().__init__(mlp)
        self.block_dict["edge_aggregator"] = edge_aggregator
        self.block_dict["node_aggregator"] = node_aggregator
        self._independent = False

    def forward(self, *, global_attr, node_attr, edge_attr, edges, node_idx, edge_idx):
        aggregated = [global_attr]
        if "node_aggregator" in self.block_dict:
            aggregated.append(
                self.block_dict["node_aggregator"](
                    node_attr, node_idx, dim=0, dim_size=global_attr.shape[0]
                )
            )
        if "edge_aggregator" in self.block_dict:
            aggregated.append(
                self.block_dict["edge_aggregator"](
                    edge_attr, edge_idx, dim=0, dim_size=global_attr.shape[0]
                )
            )

        out = torch.cat(aggregated, dim=1)
        return self.block_dict["mlp"](out)

    def forward_from_data(self, data: GraphBatch):
        return self(data.g, data.x, data.e, data.edges, data.node_idx, data.edge_idx)



import random
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Optional

from omegaconf import MISSING

from .tools import ConfigObj


class IntNumber(ABC):
    """Abstract class castable to integer."""

    @abstractmethod
    def __int__(self):
        pass


class FloatNumber(ABC):
    """Abstract class castable to float."""

    @abstractmethod
    def __float__(self):
        pass


@dataclass
class Distribution(ConfigObj, IntNumber, FloatNumber):
    low: Any = MISSING
    high: Optional[Any] = None
    name: str = MISSING


@dataclass
class Uniform(Distribution):
    """Uniform distribution class. Automatically type casts to float or int.

    .. code-block::

        x = Uniform(1, 10)
        for _ in range(10):
            print(float(x))
            print(int(x))

    .. code-block::

        x = Uniform(1.1)
        for _ in range(10):
            assert int(x) == 1
            assert float(x) == 1.1
    """

    low: float = 1
    high: Optional[float] = None
    name: str = "uniform"

    def validate(self):
        if self.high is not None:
            assert self.high >= self.low

    def __call__(self) -> float:
        if self.high is None:
            return self.low
        return random.random() * (self.high - self.low) + self.low

    def __float__(self):
        return self()

    def __int__(self):
        return int(float(self))


@dataclass
class DiscreteUniform(Uniform):
    """Discrete uniform distribution class. Automatically type casts to float
    or int.

    .. code-block::

        x = Uniform(1, 10)
        for _ in range(10):
            print(float(x))
            print(int(x))

    .. code-block::

        x = Uniform(1.1)
        for _ in range(10):
            assert int(x) == 1
            assert float(x) == 1.
    """

    low: int = 1
    high: Optional[int] = None
    name: str = "discrete_uniform"

    def __call__(self) -> int:
        if self.high is None:
            return self.low
        return random.randint(self.low, self.high)

    def __int__(self):
        return self()

    def __float__(self):
        return float(int(self))


@dataclass
class DataGenConfig(ConfigObj):
    num_graphs: int = MISSING
    density: Uniform = field(default_factory=Uniform)
    num_nodes: DiscreteUniform = field(default_factory=DiscreteUniform)
    path_length: DiscreteUniform = field(default_factory=DiscreteUniform)
    composition_density: Uniform = field(default_factory=Uniform)
    batch_size: int = MISSING
    shuffle: bool = MISSING
    name: str = ""


@dataclass
class DataConfig(ConfigObj):
    eval: DataGenConfig = field(default_factory=lambda: DataGenConfig(name="eval"))
    train: DataGenConfig = field(default_factory=lambda: DataGenConfig(name="train"))



from __future__ import annotations

from typing import List
from typing import Optional
from typing import Type

import networkx as nx
import torch

from caldera.data._graph_data import GraphData
from caldera.utils import reindex_tensor
from caldera.utils import stable_arg_sort_long
from caldera.utils import torch_scatter_group
from caldera.utils.nx.types import DirectedGraph


class GraphBatch(GraphData):
    __slots__ = GraphData.__slots__ + ["node_idx", "edge_idx"]

    # TODO: global_idx
    def __init__(
        self,
        node_attr: torch.FloatTensor,
        edge_attr: torch.FloatTensornsor,
        global_attr: torch.FloatTensor,
        edges: torch.LongTensor,
        node_idx: torch.LongTensor,
        edge_idx: torch.LongTensor,
    ):
        super().__init__(node_attr, edge_attr, global_attr, edges)
        self.node_idx = node_idx
        self.edge_idx = edge_idx
        GraphBatch.debug(self)

    @staticmethod
    def _same(a):
        return min(a) == max(a)

    def debug(self):
        super().debug()
        if not self.node_idx.dtype == torch.long:
            raise RuntimeError(
                "Wrong tensor type. `node_idx` must be dtype={} not {}".format(
                    self.node_idx.dtype, torch.long
                )
            )
        if not self.edge_idx.dtype == torch.long:
            raise RuntimeError(
                "Wrong tensor type. `edge_idx` must be dtype={} not {}".format(
                    self.edge_idx.dtype, torch.long
                )
            )
        if (
            self.node_idx.shape[0]
            and self.edge_idx.shape[0]
            and self.edge_idx.max() > self.node_idx.max()
        ):
            raise RuntimeError(
                "Number of graphs in node_idx {} and edge_idx {} mismatch".format(
                    self.node_idx.max(), self.edge_idx.max()
                )
            )
        if self.node_idx.shape[0] and self.node_idx.shape[0] != self.x.shape[0]:
            raise RuntimeError(
                "Number of node indices {} must match number of node attr {}".format(
                    self.node_idx.shape[0], self.x.shape[0]
                )
            )
        if self.edge_idx.shape[0] and self.edge_idx.shape[0] != self.e.shape[0]:
            raise RuntimeError(
                "Number of node indices {} must match number of node attr {}".format(
                    self.edge_idx.shape[0], self.e.shape[0]
                )
            )
        # if not self.node_idx.min() == 0:
        #     raise RuntimeError(
        #         "Minimum graph index (node_idx.min()) must start at 0, not {}".format(self.node_idx.min()))
        # if not self.edge_idx.min() == 0:
        #     raise RuntimeError(
        #         "Minimum graph index (edge_idx.min()) must start at 0, not {}".format(self.edge_idx.min()))

    @classmethod
    def from_data_list(cls, data_list):
        # checks
        n_features = []
        e_features = []
        g_features = []
        for data in data_list:
            n_features.append(data.x.shape[1])
            e_features.append(data.e.shape[1])
            g_features.append(data.g.shape[0])
        if not cls._same(n_features):
            raise RuntimeError("Node feature dimensions must all be the same")
        if not cls._same(e_features):
            raise RuntimeError("Edge feature dimensions must all be the same")
        if not cls._same(g_features):
            raise RuntimeError("Global feature dimensions must all be the same")

        device = data_list[0].x.device

        node_repeats = torch.tensor(
            [data.x.shape[0] for data in data_list], device=device
        )
        edge_repeats = torch.tensor(
            [data.e.shape[0] for data in data_list], device=device
        )
        node_idx = torch.repeat_interleave(
            torch.arange(0, node_repeats.shape[0], dtype=torch.long, device=device),
            node_repeats,
        )
        edge_idx = torch.repeat_interleave(
            torch.arange(0, edge_repeats.shape[0], dtype=torch.long, device=device),
            edge_repeats,
        )

        # concatenate edges
        edges = torch.cat([data.edges for data in data_list], dim=1)

        # cumulated shapes
        c = torch.cumsum(
            torch.tensor(
                [0] + [data.x.shape[0] for data in data_list[:-1]], device=device
            ),
            dim=0,
        )
        delta = torch.repeat_interleave(c, edge_repeats).repeat(2, 1)

        # shift concatenated edges
        edges = edges + delta

        return cls(
            node_attr=torch.cat([data.x for data in data_list]),
            edge_attr=torch.cat([data.e for data in data_list]),
            global_attr=torch.cat([data.g for data in data_list]),
            edges=edges,
            node_idx=node_idx,
            edge_idx=edge_idx,
        )

    def to_data_list(self) -> List[GraphData]:
        gidx_n, node_attr = torch_scatter_group(self.x, self.node_idx)
        gidx_e, edge_attr = torch_scatter_group(self.e, self.edge_idx)
        gidx_edge, edges = torch_scatter_group(self.edges.T, self.edge_idx)

        def to_dict(a, b):
            return dict(zip([x.item() for x in a], b))

        ndict = to_dict(gidx_n, node_attr)
        edict = to_dict(gidx_e, edge_attr)
        edgesdict = to_dict(gidx_edge, edges)
        datalist = []
        n_sizes = 0
        for k in ndict:
            _edges = edgesdict[k].T
            _edges -= n_sizes
            n_sizes += ndict[k].shape[0]

            data = GraphData(ndict[k], edict[k], torch.unsqueeze(self.g[k], 0), _edges)
            datalist.append(data)
        return datalist

    def to_networkx(self, *args, **kwargs):
        raise NotImplementedError(
            "Method not implemented for batches. Use `{}`".format(
                self.to_networkx_list.__name__
            )
        )

    def from_networkx(self, *args, **kwargs):
        raise NotImplementedError(
            "Method not implemented for batches. Use `{}`".format(
                self.from_networkx_list.__name__
            )
        )

    def to_networkx_list(
        self,
        feature_key: str = "features",
        global_attr_key: str = "data",
        graph_type: Type[DirectedGraph] = nx.OrderedMultiDiGraph,
    ) -> List[DirectedGraph]:
        graphs = []
        for data in self.to_data_list():
            graphs.append(
                data.to_networkx(
                    feature_key, global_attr_key=global_attr_key, graph_type=graph_type
                )
            )
        return graphs

    @staticmethod
    def from_networkx_list(
        graphs: List[DirectedGraph], *args, **kwargs
    ) -> "GraphBatch":
        data_list = [GraphData.from_networkx(g, *args, **kwargs) for g in graphs]
        return GraphBatch.from_data_list(data_list)

    def append_nodes(self, node_attr: torch.Tensor, node_idx: torch.Tensor):
        datalist = self.to_data_list()
        idx, groups = torch_scatter_group(node_attr, node_idx)
        for i, g in zip(idx, groups):
            data = datalist[i.item()]
            data.append_nodes(g)
        batch = GraphBatch.from_data_list(datalist)
        self.x = batch.x
        self.e = batch.e
        self.g = batch.g
        self.node_idx = batch.node_idx
        self.debug()
        return self

    def append_edges(
        self, edge_attr: torch.Tensor, edges: torch.Tensor, edge_idx: torch.Tensor
    ):
        """Append edges to the graph batch at the specified edge_idx (assumed
        to be sorted).

        :param edge_attr:
        :param edges:
        :param edge_idx:
        :return:
        """
        edges = torch.cat([self.edges, edges], dim=1)
        e = torch.cat([self.e, edge_attr])
        edge_idx = torch.cat([self.edge_idx, edge_idx])
        i = stable_arg_sort_long(edge_idx)

        self.edges = edges[:, i]
        self.e = e[i]
        self.edge_idx = edge_idx[i]
        self.debug()
        return self

    def _mask_dispatch(
        self,
        node_mask: Optional[torch.BoolTensor],
        edge_mask: Optional[torch.BoolTensor],
        as_view: bool,
        detach: bool,
        new_inst: bool,
    ):
        self._validate_masks(node_mask, edge_mask)
        edges = self._apply_mask(self.edges, edge_mask, detach, as_view, dim=1)
        edges = self._mask_dispatch_reindex_edges(edges, node_mask)
        x = self._apply_mask(self.x, node_mask, detach, as_view)
        e = self._apply_mask(self.e, edge_mask, detach, as_view)
        g = self._apply_mask(self.g, None, detach, as_view)
        node_idx = self._apply_mask(self.node_idx, node_mask, detach, as_view)
        edge_idx = self._apply_mask(self.edge_idx, edge_mask, detach, as_view)

        node_idx, edge_idx = reindex_tensor(node_idx, edge_idx)

        return self._mask_dispatch_constructor(
            new_inst, x, e, g, edges, node_idx=node_idx, edge_idx=edge_idx
        )

    # def append_edges
    # def collect_and_collate(x1, i1, x2, i2, collate_fn = torch.cat):
    #     i1, groups1 = scatter_group(x1, i1)
    #     i2, groups2 = scatter_group(x2, i2)
    #     d1 = {k.item(): v for k, v in zip(i1, groups1)}
    #     d2 = {k.item(): v for k, v in zip(i2, groups2)}
    #     return dict_collate(d1, d2, torch.cat)
    #
    #
    # i = stable_arg_sort_long(edge_idx)
    # edge_attr = edge_attr[i]
    # edges = edges[:, i]
    #
    # d = collect_and_collate(edge_attr, edge_idx, self.e, self.edge_idx)
    # keys = sorted(d)
    # e = torch.cat([d[k] for k in keys])
    #
    # d = collect_and_collate(edges.T, edge_idx, self.edges.T, self.edge_idx)
    # keys = sorted(d)
    # edges = torch.cat([d[k] for k in keys], dim=0)
    #
    # d = collect_and_collate(edge_idx, edge_idx, self.edge_idx, self.edge_idx)
    # keys = sorted(d)
    # edge_idx = torch.cat([d[k] for k in keys], dim=0)
    #
    # self.e = e
    # self.edges = edges.T
    # self.edge_idx = edge_idx
    # self.debug()
    # return self
    # e = torch.cat([d[k] for k in keys])

    #
    # def collate_shape(x):
    #     return [[i] * _x.shape[0] for i, _x in enumerate(x)]
    #
    # d = dict_collate(d1, d2, collate_shape)
    # u = [torch.Tensor(functools.reduce(operator.add, v)) for v in d.values()]
    # edge_idx = torch.cumsum(torch.cat(u), 0)
    #
    # self.e = e
    # self.edge_idx = edge_idx
    # self.debug()
    # return self

    # i1, g1 = scatter_group(self.x, self.node_idx)
    # i2, g2 = scatter_group(node_attr, node_idx)
    # d1 = {k.item(): v for k, v in zip(i1, g1)}
    # d2 = {k.item(): v for k, v in zip(i2, g2)}
    #
    # # collate
    # d = dict_collate(d1, d2, torch.cat)
    # keys = sorted(d.keys())
    # x = torch.cat([d[k] for k in keys])
    #
    # # collect node indices
    # node_idx = []
    # for k in keys:
    #     node_idx += [k] * d[k].shape[0]
    #
    # # correct edges
    # def get_shape(x):
    #     return [[i] * _x.shape[0] for i, _x in enumerate(x)]
    #
    # u = [torch.Tensor(functools.reduce(operator.add, v)) for v in dict_collate(d1, d2, get_shape).values()]
    # delta_edges = torch.cumsum(torch.cat(u), 0)
    # self.edges += delta_edges
    # self.x = x
    # self.node_idx = torch.tensor(node_idx, dtype=torch.long)
    # self.debug()
    #
    # def append_nodes(self, edge_attr: torch.Tensor, edge_idx: torch.Tensor):
    #     i1, g1 = scatter_group(self.x, self.node_idx)
    #     i2, g2 = scatter_group(edge_attr, node_idx)
    #     d1 = {k.item(): v for k, v in zip(i1, g1)}
    #     d2 = {k.item(): v for k, v in zip(i2, g2)}
    #
    #     # collate
    #     d = dict_collate(d1, d2, torch.cat)
    #     keys = sorted(d.keys())
    #     x = torch.cat([d[k] for k in keys])
    #
    #     # collect node indices
    #     node_idx = []
    #     for k in keys:
    #         node_idx += [k] * d[k].shape[0]
    #
    #     self.x = x
    #     self.node_idx = torch.tensor(node_idx, dtype=torch.long)
    #     self.debug()

    def _eq_helper(self, *args, **kwargs):
        raise NotImplementedError("Cannot compare batches")

    def allclose(self, *args, **kwargs):
        raise NotImplementedError("Cannot compare batches")

    @classmethod
    def random(cls, *args, **kwargs):
        raise NotImplementedError("Not implemented for {}".format(cls))

    @classmethod
    def random_batch(
        cls,
        size: int,
        n_feat: int,
        e_feat: int,
        g_feat: int,
        min_nodes: int = 1,
        max_nodes: int = 10,
        min_edges: int = 1,
        max_edges: int = 20,
        requires_grad: Optional[bool] = None,
    ) -> GraphBatch:
        datalist = []
        for _ in range(size):
            datalist.append(
                GraphData.random(
                    n_feat,
                    e_feat,
                    g_feat,
                    min_nodes=min_nodes,
                    max_nodes=max_nodes,
                    min_edges=min_edges,
                    max_edges=max_edges,
                )
            )
        batch = cls.from_data_list(datalist)
        if requires_grad is not None:
            batch.requires_grad = requires_grad
        return batch

    def __eq__(self, *args, **kwargs):
        raise NotImplementedError("Cannot compare batches")

    def view(
        self,
        x_slice: Optional[slice] = None,
        e_slice: Optional[slice] = None,
        g_slice: Optional[slice] = None,
        edges_slice: Optional[slice] = None,
    ) -> GraphData:
        if x_slice is None:
            x_slice = slice(None, None, None)
        if e_slice is None:
            e_slice = slice(None, None, None)
        if g_slice is None:
            g_slice = slice(None, None, None)
        if edges_slice is None:
            edges_slice = slice(None, None, None)
        return self.__class__(
            self.x[:, x_slice],
            self.e[:, e_slice],
            self.g[:, g_slice],
            self.edges[:, edges_slice],
            self.node_idx[:],
            self.edge_idx[e_slice],
        )

    def disjoint_union(self, other: GraphBatch) -> GraphBatch:
        """Disjoint union between two GraphBatches.

        :param other:
        :return:
        """
        x = torch.cat(self.x, other.x)
        e = torch.cat(self.e, other.e)
        g = torch.cat(self.g, other.g)

        n = self.node_idx.max()
        node_idx = torch.cat(self.node_idx, other.node_idx + n)
        edge_idx = torch.cat(self.edge_idx, other.edge_idx + n)
        edges = torch.cat(self.edges, other.edges + n)
        node_idx, edge_idx, edges = reindex_tensor(node_idx, edge_idx, edges)

        return GraphBatch(x, e, g, edges, node_idx, edge_idx)

    def shuffle_graphs_(self) -> None:
        b = torch.unique(self.node_idx)
        ridx = torch.randperm(b.shape[0])
        self.g = self.g[ridx]
        _, node_idx, edge_idx = reindex_tensor(ridx, self.node_idx, self.edge_idx)
        self.node_idx = node_idx
        self.edge_idx = edge_idx

    def shuffle_graphs(self) -> GraphBatch:
        cloned = self.detach().clone()
        cloned.shuffle_graphs_()
        return cloned

    def shuffle_(self) -> None:
        self.shuffle_graphs_()
        self.shuffle_nodes_()
        self.shuffle_edges_()

    def shuffle(self) -> GraphBatch:
        cloned = self.detach().clone()
        cloned.shuffle_()
        return cloned



import functools
import itertools
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Hashable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

import networkx as nx
import numpy as np
from networkx.classes.reportviews import EdgeView
from networkx.classes.reportviews import NodeView

from ._base import NetworkxTransformBase
from ._types import _G
from caldera.utils import dict_join
from caldera.utils import functional as fn
from caldera.utils.tensor import to_one_hot

GraphGen = Generator[_G, None, None]

_T = TypeVar("T")
_K = TypeVar("K")
_S = TypeVar("S")
_V = TypeVar("V")


def values_to_one_hot(
    values: Iterable[_T],
    classes: Union[List[_T], Tuple[_T, ...]],
    num_classes: Optional[int] = None,
) -> np.ndarray:
    """Convert an iterable of values to a one-hot encoded `np.ndarray`.

    :param values: iterable of values
    :param classes: valid classes of values. Will appear in one-hot array in order they appear here.
    :param num_classes: Number of classes for one-hot encoder. If not provided, number of provided classes
        will be used.
    :return: one-hot encoded array of values.
    """
    assert len(set(classes)) == len(classes)
    d = {k: i for i, k in enumerate(classes)}
    _values = []
    for v in values:
        try:
            _values.append(d[v])
        except KeyError:
            raise KeyError(
                "Value '{}' not found in list of available one-hot classes: {}".format(
                    v, d
                )
            )
    if num_classes is None:
        num_classes = len(d)
    return to_one_hot(np.array(_values), mx=num_classes)


# def _get_join_fn(join_func):
#     if join_func == "hstack":
#
#         def join_func(a, b):
#             return np.hstack([a, b])
#
#     elif join_func == "vstack":
#
#         def join_func(a, b):
#             return np.vstack([a, b])
#
#     return join_func


# def _get_processing_func(encoding):
#     if encoding is None:
#         processing_func = fn.compose(list, np.array)
#     elif encoding == "onehot":
#         processing_func = fn.compose(list, functools.partial(values_to_one_hot))
#     else:
#         processing_func = None
#     return processing_func


def _dispatch_nx_iterator(
    g: nx.DiGraph, x: str
) -> Union[NodeView, EdgeView, Generator[Tuple[Hashable, Any], None, None]]:
    if x == "node":
        return g.nodes(data=True)
    elif x == "edge":
        return g.edges(data=True)
    elif x == "global":
        return g.globals(data=True)
    else:
        raise ValueError("choose from {}".format(["node", "edge", "global"]))


def _update_left_inplace(
    data: Dict[_K, _T], new_data: Dict[_K, _S], join_fn: Callable[[_T, _S], _K]
) -> Dict[_K, Union[_T, _S, _V]]:
    """Updates the left dictionary, joining values with the provided
    function."""
    return dict_join(data, new_data, data, mode="right", join_fn=join_fn)


def _merge_update(data, key, to_key, join_fn, process_fn, default=...):
    """Update dictionary by applying join_fn and process_fn. For example, the
    processing function may be converting values to np.ndarray, while the
    join_fn might be to apply hstack.

    :param data:
    :param key:
    :param to_key:
    :param join_fn:
    :param process_fn:
    :param default:
    :return:
    """

    select_data = fn.index_each(-1)

    getter = fn.get_each(key)
    if default is not ...:
        getter = fn.get_each(key, default=default)

    select_process_and_send_to_key = fn.compose(
        fn.index_each(-1),
        getter,
        process_fn,
        fn.map_each(lambda x: {to_key: x}),
    )
    merge_and_join = functools.partial(_update_left_inplace, join_fn=join_fn)

    teed = itertools.tee(data)
    original_data = select_data(teed[0])
    processed_data = list(select_process_and_send_to_key(teed[1]))

    for d1, d2 in zip(original_data, processed_data):
        merge_and_join(d1, d2)


class NetworkxAttachNumpyFeatures(NetworkxTransformBase):
    def __init__(
        self,
        x: str,
        from_key: str,
        to_key: str,
        *,
        default: Any = ...,
        encoding: Optional[Union[str, Callable[[Iterable[_T]], Iterable[_T]]]] = None,
        join_func: Union[
            str, Callable[[np.ndarray, np.ndarray], np.ndarray]
        ] = "hstack",
        global_key: str = None,
        **processing_kwargs
    ):
        """Initialize transform that converts networkx features to a
        :class:`np.ndarray`

        :param x: 'edge', 'node', or 'global'
        :param from_key: dictionary key to find feature
        :param to_key: new key to attach feature
        :param default:
        :param encoding:
        :param join_func: select from 'hstack', 'vstack' or provide a new join function
        :param global_key:
        :param processing_kwargs:
        """

        if encoding is None:
            processing_func = fn.compose(list, np.array)
        elif encoding == "onehot":
            processing_func = fn.compose(
                list, functools.partial(values_to_one_hot, **processing_kwargs)
            )
        elif encoding == "bool":
            processing_func = fn.compose(
                fn.map_each(lambda x: np.array([int(bool(x))]))
            )
        else:
            raise ValueError(
                "Encoding {} is not a valid encoding. Select from 'onehot' or None".format(
                    encoding
                )
            )

        if join_func == "hstack":

            def join_func(a, b):
                return np.hstack([a, b])

        elif join_func == "vstack":

            def join_func(a, b):
                return np.vstack([a, b])

        self.x = x
        self.from_key = from_key
        self.to_key = to_key
        self.global_key = global_key
        self.default = default
        self.join_fn = join_func
        self.processing_func = processing_func
        self.processing_kwargs = processing_kwargs

    def transform(self, g: _G) -> _G:
        iterator = list(_dispatch_nx_iterator(g, self.x))

        _merge_update(
            iterator,
            self.from_key,
            self.to_key,
            join_fn=self.join_fn,
            process_fn=self.processing_func,
            default=self.default,
        )
        return g


class NetworkxAttachNumpyOneHot(NetworkxAttachNumpyFeatures):
    def __init__(
        self,
        x: str,
        from_key: str,
        to_key: str,
        *,
        default: Any = ...,
        join_func: str = "hstack",
        global_key: str = None,
        classes: List[str] = None,
        num_classes: int = None
    ):
        """Initialize transform that converts encodes networkx features into
        one-hot encodings.

        :param x: 'edge', 'node', 'global'
        :param from_key: dictionary key to find feature
        :param to_key: new key to attach feature
        :param default:
        :param join_func:
        :param global_key:
        :param classes:
        :param num_classes:
        """
        super().__init__(
            x,
            from_key,
            to_key,
            default=default,
            encoding="onehot",
            join_func=join_func,
            global_key=global_key,
            classes=classes,
            num_classes=num_classes,
        )


class NetworkxAttachNumpyBool(NetworkxAttachNumpyFeatures):
    def __init__(
        self,
        x: str,
        from_key: str,
        to_key: str,
        *,
        default: Any = ...,
        join_func: str = "hstack",
        global_key: str = None,
        classes: List[str] = None,
        num_classes: int = None
    ):
        """Initialize transform that converts encodes networkx into a boolean
        represented as :class:`np.ndarray` of size 1.

         E.g. `True` gets encoded as `np.ndarray([1.])`

        :param x: 'edge', 'node', 'global'
        :param from_key: dictionary key to find feature
        :param to_key: new key to attach feature
        :param default:
        :param join_func:
        :param global_key:
        :param classes:
        :param num_classes:
        """
        super().__init__(
            x,
            from_key,
            to_key,
            default=default,
            encoding="bool",
            join_func=join_func,
            global_key=global_key,
            classes=classes,
            num_classes=num_classes,
        )



import inspect
import re
from multiprocessing.pool import IMapIterator
from multiprocessing.pool import Pool
from typing import Callable
from typing import List
from typing import Tuple
from typing import TypeVar

import dill

_T = TypeVar("T")
_S = TypeVar("S")
_Arg = TypeVar("A")


def run_payload(payload: bytes) -> Tuple[int, _T]:
    unloaded: Tuple[Callable[[_Arg], _T], _Arg, int] = dill.loads(payload)
    fun, args, idx = unloaded
    return idx, fun(args)


def create_payloads(func: Callable[[_Arg], _T], args: List[_Arg]) -> List[bytes]:
    payloads = []
    for idx, arg in enumerate(args):
        data = (func, arg, idx)
        payloads.append(dill.dumps(data))
    return payloads


def handle_result(
    results: IMapIterator, callback=None, error_callback=None
) -> List[Tuple[int, _T]]:
    results_list = []
    while True:
        r = ...
        try:
            idx, r = next(results)
        except StopIteration:
            break
        except Exception as e:
            if error_callback:
                error_callback(e)
            else:
                raise e
        finally:
            if r is not ...:
                if callback:
                    callback(r)
                results_list.append((idx, r))
    return results_list


def _identity_wrapper(f: Callable[[_S], _T]) -> Callable[[_Arg], _T]:
    return f


def star_wrapper(f: Callable[[_S], _T]) -> Callable[[_Arg], _T]:
    def wrapped(args):
        return f(*args)

    return wrapped


def _resolved_sorted_handler(resolved: List[Tuple[int, _T]]) -> List[_T]:
    return [r[1] for r in sorted(resolved)]


_default_wrapper = _identity_wrapper
_default_resolved_handler = _identity_wrapper


def run_with_pool(
    pool: Pool,
    func: Callable,
    args: List[_Arg],
    wrapper: Callable[[Callable[[_S], _T]], Callable[[_Arg], _T]] = None,
    chunksize: int = 1,
    callback: Callable[[_T], None] = None,
    error_callback: Callable[[Exception], None] = None,
    resolved_handler: Callable[[int, _T], _S] = None,
) -> _S:
    wrapper = wrapper or _default_wrapper
    payloads = create_payloads(wrapper(func), args)
    results = pool.imap_unordered(run_payload, payloads, chunksize=chunksize)
    resolved: List[Tuple[int, _T]] = handle_result(results, callback, error_callback)

    resolved_handler = resolved_handler or _default_resolved_handler
    return resolved_handler(resolved)


def argmap(f, args, kwargs):
    argspec = inspect.getfullargspec(f)
    mapping = {}
    for arg, val in zip(argspec.args, args):
        mapping[arg] = val
    mapping.update(kwargs)
    return mapping


_varnamepattern = re.compile(r"[a-zA-Z_][\w\d]*")


def valid_varname(x: str):
    if not isinstance(x, str):
        return False
    if _varnamepattern.match(x) is not None:
        return True
    return False



import operator
from functools import reduce
from typing import overload
from typing import Tuple
from typing import Type
from typing import Union

import torch

SizeType = Union[torch.Size, Tuple[int, ...]]


def prod(x):
    if not x:
        return 1
    else:
        return reduce(operator.mul, x)


def repeat_roll(shape: SizeType, dim: int) -> torch.Tensor:
    """Roll over shape, rolling the repeat over different dimension."""
    torch.arange(shape[dim])

    repeat_dims = list(shape)[:dim]
    interleave_dims = list(shape[dim + 1 :])

    n = prod(interleave_dims)
    m = prod(repeat_dims)
    return torch.repeat_interleave(torch.arange(shape[dim]), n).repeat(m)


def unroll_index(
    shape: SizeType, dtype: Type = torch.long
) -> Tuple[torch.LongTensor, ...]:
    idx = tuple()
    for dim in range(len(shape)):
        idx = idx + (repeat_roll(shape, dim).to(dtype),)
    return idx


@overload
def unravel_index(index: int, shape: ...) -> Tuple[int]:
    ...


def unravel_index(
    index: torch.LongTensor, shape: SizeType
) -> Tuple[torch.LongTensor, ...]:
    out = []
    for dim in reversed(shape):
        out.append(index % dim)
        index = index // dim
    return tuple(reversed(out))


@overload
def reindex_tensor(a: torch.Tensor) -> torch.Tensor:
    ...


def reindex_tensor(
    a: torch.Tensor, *tensors: Tuple[torch.Tensor, ...]
) -> Tuple[torch.Tensor, ...]:
    """Reindex a tensor to lowest index. Handles multiple tensors and tensors
    with many dimensions.

    .. code-block::

        a = torch.tensor([1, 1, 1, 4, 0, 5, 0, 0, 0])
        b = reindex(a)
        print(b)
        # >> tensor([0, 0, 0, 1, 2, 3, 2, 2, 2])

    .. code-block::

        # multiple tensors with multiple dimensions
        a = torch.tensor([1, 1, 1, 1, 0, 2, 0, 5, 6])
        b = torch.tensor([[6, 5, 1, 70], [0, 80, 5, 6]])
        expected1 = torch.tensor([0, 0, 0, 0, 1, 2, 1, 3, 4])
        expected2 = torch.tensor([[4, 3, 0, 5], [1, 6, 3, 4]])
        c, d = reindex_tensor(a, b)
        assert torch.all(c == expected1)
        assert torch.all(d == expected2)


    :param a: tensor to reindex
    :return: new reindexed tensor
    """

    values_list = []
    all_tensors = [a] + list(tensors)
    for t in all_tensors:
        if not t.dtype == a.dtype:
            raise ValueError(
                "All tensors must be same type. {} != {}".format(t.dtype, a.dtype)
            )
        values_list.append(t.flatten().tolist())

    new_tensors = []

    replace = {}
    j = 0
    for tidx, tlist in enumerate(values_list):
        t = all_tensors[tidx]
        b = torch.empty_like(t)
        new_tensors.append(b)
        for i, _t in enumerate(tlist):
            if _t not in replace:
                replace[_t] = j
                j += 1
            bidx = unravel_index(i, t.shape)
            b[bidx] = replace[_t]

    if tensors:
        return new_tensors
    else:
        return new_tensors[0]



""""""
from ._version import __authors__  # noqa: F401
from ._version import __homepage__  # noqa: F401
from ._version import __repository__  # noqa: F401
from ._version import __title__  # noqa: F401
from ._version import __version__  # noqa: F401
from caldera._setup import setup as _setup

_setup()



from typing import List
from typing import Optional
from typing import overload

import torch
from torch.utils.data import Dataset

from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.transforms._base import TransformCallable


class GraphDataset(Dataset):
    """Graph dataset."""

    def __init__(
        self, datalist: List[GraphData], transform: Optional[TransformCallable] = None
    ):
        """Creates a graph dataset.

        :param datalist:
        :param transform:
        """
        assert isinstance(datalist, list)
        self.datalist = datalist
        self.transform = transform

    def __len__(self):
        return len(self.datalist)

    def mem_sizes(self):
        return torch.IntTensor([d.memsize() for d in self.dataset])

    def get(self, idx, transform: Optional[TransformCallable] = None):
        transform = transform or self.transform
        if torch.is_tensor(idx):
            idx = idx.tolist()

        if isinstance(idx, int):
            idx = [idx]
            one_item = True
        else:
            one_item = False

        if isinstance(idx, slice):
            samples = self.datalist[idx]
        else:
            samples = []
            for i in idx:
                sample = self.datalist[i]
                if transform:
                    sample = transform(sample)
                samples.append(sample)

        if one_item:
            samples = samples[0]
        return samples

    @overload
    def __getitem__(self, idx: torch.LongTensor) -> List[GraphData]:
        ...

    @overload
    def __getitem__(self, idx: List[int]) -> List[GraphData]:
        ...

    def __getitem__(self, idx: int) -> GraphData:
        return self.get(idx)


class GraphBatchDataset(GraphDataset):
    """Dataset that loads a :class:`caldera.data.GraphData` list into.

    :class:`caldera.data.GraphBatch` instances, and then performs transform (if
    provided).
    """

    @overload
    def __getitem__(self, idx: torch.LongTensor) -> GraphBatch:
        ...

    @overload
    def __getitem__(self, idx: List[int]) -> GraphData:
        ...

    def __getitem__(self, idx: int) -> GraphBatch:
        samples = self.get(idx, transform=None)
        if isinstance(samples, GraphData):
            samples = [samples]
        batch = GraphBatch.from_data_list(samples)
        if self.transform:
            batch = self.transform(batch)
        return batch



from typing import Callable
from typing import Dict
from typing import TypeVar
from typing import Union

from ._nx_feature_transform import NetworkxTransformFeatureData


T = TypeVar("T")
S = TypeVar("S")
K = TypeVar("K")


class NetworkxApplyToKey(NetworkxTransformFeatureData):
    def __init__(
        self,
        key: str,
        node_func: Callable[[S], T] = None,
        edge_func: Callable[[S], T] = None,
        glob_func: Callable[[S], T] = None,
    ):
        """Transformation to apply a function to the a keyed value for a]
        networkx graph.

        :param key: key to find the value to apply the function to.
        :param node_func:
        :param edge_func:
        :param glob_func:
        """

        def apply_to_key(f: Callable[[S], T]) -> Union[Callable[[Dict[K, S]], T], None]:
            if f is None:
                return None

            def _apply_to_key(d: Dict[K, S]) -> T:
                d[key] = f(d[key])
                return d

            return _apply_to_key

        super().__init__(
            node_transform=apply_to_key(node_func),
            edge_transform=apply_to_key(edge_func),
            global_transform=apply_to_key(glob_func),
        )



from typing import Optional
from typing import Type

from ._base import NetworkxTransformBase
from caldera.utils.nx import nx_to_directed


class NetworkxToDirected(NetworkxTransformBase):
    def __init__(self, graph_class: Optional[Type] = None):
        self.graph_class = graph_class

    def transform(self, g):
        if self.graph_class is None:
            kwargs = {}
        else:
            kwargs = dict(graph_class=self.graph_class)
        return nx_to_directed(g, **kwargs)



# g = random_graph((100, 150), d=(0.01, 0.03), e=None)
# annotate_shortest_path(g)
# # nx.draw(g)
# pos = nx.layout.spring_layout(g)
# nodelist = list(g.nodes)
# node_color = []
# for n in nodelist:
#     node_color.append(g.nodes[n]['target'][0])
# edge_list = []
# edge_color = []
# for n1, n2, edata in g.edges(data=True):
#     edge_list.append((n1, n2))
#     edge_color.append(edata['target'][0])
# print(node_color)
# nx.draw_networkx_edges(g, pos=pos, width=0.5, edge_color=edge_color)
# nx.draw_networkx_nodes(g, pos=pos, node_color=node_color, node_size=10)



# flake8: noqa
import collections

import networkx as nx
import numpy as np
import pylab as plt


def sorted_center(iterable, reverse=False, key=None, select=None):
    centered = []
    s = sorted(iterable, key=key, reverse=reverse)
    front = False
    for v in s:
        if not front:
            centered.append(v)
        else:
            centered.insert(0, v)
        front = not front
    if select:
        return [select(_x) for _x in centered]
    return centered


class GraphPlotter:
    def __init__(self, graph, ax=None, pos=None):
        if ax is None:
            ax = plt.figure(figsize=(3, 3)).gca()
            ax.axis("off")
            ax.set_xlim(0, 1.0)
            ax.set_ylim(0, 1.0)

        self._ax = ax
        self._graph = graph
        if pos is None:
            self._pos = dict()
            self.spring_layout()
        else:
            self._pos = pos

    @property
    def _base_draw_kwargs(self):
        return dict(G=self._graph, ax=self._ax, pos=self._pos)

    def topological_sort(self):
        xlim = self._ax.get_xlim()
        ylim = self._ax.get_ylim()

        x = xlim[1] - xlim[0]
        y = ylim[1] - ylim[0]
        xspacer = x * 0.05
        yspacer = y * 0.05

        new_xlim = (xlim[0] + xspacer, xlim[1] - xspacer)
        new_ylim = (ylim[0] + yspacer, ylim[1] - yspacer)
        pos = self._topological_sort(self._graph, xlim=new_xlim, ylim=new_ylim)
        self._pos = pos

    @staticmethod
    def _roots_and_leaves(G, include_cycles=False):
        roots = [k for k, v in G.in_degree(G.nodes) if v == 0]
        leaves = [k for k, v in G.out_degree(G.nodes) if v == 0]

        if include_cycles:
            for c in nx.simple_cycles(G):
                outdegree = list(G.out_degree(c))
                outdegree.sort(key=lambda x: x[1])
                leaves.append(outdegree[0][0])

                indegree = list(G.out_degree(c))
                indegree.sort(key=lambda x: x[1])
                roots.append(indegree[0][0])
        return roots, leaves

    @classmethod
    def _get_roots(cls, G, include_cycles=False):
        return cls._roots_and_leaves(G, include_cycles=include_cycles)[0]

    @classmethod
    def _get_leaves(cls, G, include_cycles=False):
        return cls._roots_and_leaves(G, include_cycles=include_cycles)[1]

    @staticmethod
    def _get_subgraphs(graph):
        """Get independent subgraphs."""
        node_list = list(graph.nodes)
        subgraphs = []
        while len(node_list) > 0:
            node = node_list[-1]
            subgraph = nx.bfs_tree(to_undirected(graph), node)
            for n in subgraph.nodes:
                node_list.remove(n)
            subgraphs.append(graph.subgraph(subgraph.nodes))
        return subgraphs

    @classmethod
    def _topological_sort(cls, G, xlim=None, ylim=None):
        if xlim is None:
            xlim = [0.05, 0.95]
        if ylim is None:
            ylim = [0.05, 0.95]

        rts = cls._get_roots(G, include_cycles=True)
        max_depth = {}
        for root in rts:
            depths = nx.single_source_shortest_path_length(G, root)
            for n, d in depths.items():
                max_depth[n] = max(max_depth.get(n, d), d)

        by_depth = collections.OrderedDict()

        for node, depth in max_depth.items():
            by_depth.setdefault(depth, [])
            by_depth[depth].append(node)

        # center nodes with highest degree
        for depth, nodes in by_depth.items():
            centered = sorted_center(
                list(G.degree(nodes)),
                key=lambda x: x[1],
                reverse=True,
                select=lambda x: x[0],
            )
            by_depth[depth] = centered

        # push roots 'up' so they are not stuck on layer one
        for root in rts:
            successors = list(G.successors(root))
            if len(successors) > 0:
                min_depth = min([max_depth[s] for s in successors])
                max_depth[root] = min_depth - 1

        # assign positions

        y_min_max = xlim
        x_min_max = ylim
        max_width = max([len(layer) for layer in by_depth.values()])

        y_step = (y_min_max[1] - y_min_max[0]) / (max(by_depth.keys()) + 1)
        x_step = (x_min_max[1] - x_min_max[0]) / (max_width)
        positions = {}

        for depth in sorted(by_depth):
            y = y_step * depth + y_min_max[0]
            node_ids = by_depth[depth]
            w = len(node_ids)
            delta_w = max_width - w
            for i, n in enumerate(node_ids):
                x_offset = delta_w * x_step / 2.0
                x = x_min_max[0] + x_offset + i * x_step
                positions[n] = (x, y)
        return positions

    def spring_layout(self, **kwargs):
        pos = nx.spring_layout(self._graph, **kwargs)
        self._pos.update(pos)
        return self._pos

    @property
    def nodes(self):
        return self._graph.nodes

    @property
    def edges(self):
        return self._graph.edges

    def node_attrs(self):
        keys = set()
        for _, node_data in self._graph.nodes(data=True):
            keys.update(set(node_data.keys()))
        attrs = {}
        for _, node_data in self._graph.nodes(data=True):
            for k in keys:
                attrs.setdefault(k, list())
                attrs[k].append(node_data.get(k, None))
        return attrs

    def edge_attrs(self):
        keys = set()
        for _, _, edge_data in self._graph.edges(data=True):
            keys.update(set(edge_data.keys()))
        attrs = {}
        for _, _, edge_data in self._graph.edges(data=True):
            for k in keys:
                attrs.setdefault(k, list())
                attrs[k].append(edge_data.get(k, None))
        return attrs

    def map_edge_attrs(self, attrs, source, target):
        vals = [
            self._normalize(self.edge_attrs()[attr], source, target) for attr in attrs
        ]
        return dict(zip(attrs, vals))

    def map_node_attrs(self, attrs, source, target):
        vals = [
            self._normalize(self.node_attrs()[attr], source, target) for attr in attrs
        ]
        return dict(zip(attrs, vals))

    def _normalize(arr, source, target):
        x = source[1] - source[0]
        y = target[1] - target[0]
        return [_x / x * y + target[0] for _x in arr]

    def _make_draw_kwargs(self, **kwargs):
        kwargs.update(self._base_draw_kwargs)
        return kwargs

    def _draw(self, draw_function, zorder=None, **kwargs):
        draw_kwargs = self._make_draw_kwargs(**kwargs)
        collection = draw_function(**draw_kwargs)
        if collection is not None and zorder is not None:
            try:
                # This is for compatibility with older matplotlib.
                collection.set_zorder(zorder)
            except AttributeError:
                # This is for compatibility with newer matplotlib.
                collection[0].set_zorder(zorder)
        return collection

    def draw_nodes(self, **kwargs):
        """Useful kwargs: nodelist, node_size, node_color, linewidths."""
        if (
            "node_color" in kwargs
            and isinstance(kwargs["node_color"], collections.Sequence)
            and len(kwargs["node_color"]) in {3, 4}
            and not isinstance(
                kwargs["node_color"][0], (collections.Sequence, np.ndarray)
            )
        ):
            num_nodes = len(kwargs.get("nodelist", self.nodes))
            kwargs["node_color"] = np.tile(
                np.array(kwargs["node_color"])[None], [num_nodes, 1]
            )
        return self._draw(nx.draw_networkx_nodes, **kwargs)

    def draw_edges(self, **kwargs):
        """Useful kwargs: edgelist, width."""
        return self._draw(nx.draw_networkx_edges, **kwargs)

    def draw_graph(
        self,
        node_size=200,
        node_color=(0.4, 0.8, 0.4),
        node_linewidth=1.0,
        edge_width=1.0,
    ):

        node_border_color = (0.0, 0.0, 0.0, 1.0)

        # Plot nodes.
        self.draw_nodes(
            nodelist=self.nodes,
            node_size=node_size,
            node_color=node_color,
            linewidths=node_linewidth,
            edgecolors=node_border_color,
            zorder=20,
        )
        # Plot edges.
        self.draw_edges(edgelist=self.edges, width=edge_width, zorder=10)


g = nx.balanced_tree(2, 2)
g = nx.to_directed(g)

for e in g.edges:
    g.edges[e[0], e[1]]["weight"] = 1

ax = plt.figure(figsize=(3, 3)).gca()
ax.axis("off")
ax.set_xlim(0, 1.0)
ax.set_ylim(0, 1.0)

plotter = GraphPlotter(g, ax=ax)
plotter.topological_sort()

# plotter.draw_graph(node_color=(0, 0, 0), edge_width=plotter.edge_attrs()['weight'])



"""summary.py.

Methods for summaries and statistics
"""
import numpy as np
import torch

from caldera.data import GraphDataLoader


def degree(graph_batch, i):
    number_of_nodes = graph_batch.x.shape[0]
    counts = torch.unique(graph_batch.edges[i], return_counts=True)[1]
    degree = torch.cat(
        [counts, torch.zeros(number_of_nodes - counts.shape[0], dtype=torch.long)]
    ).to(torch.float)
    return degree.mean(), degree.std()


def loader_summary(loader: GraphDataLoader):
    x = []
    e = []
    g = []
    indegrees = []
    outdegrees = []
    for d, _ in loader:
        number_of_nodes = d.x.shape[0]
        number_of_edges = d.e.shape[0]
        number_of_graphs = d.g.shape[0]

        x.append(number_of_nodes)
        e.append(number_of_edges)
        g.append(number_of_graphs)

        indegrees.append(degree(d, 1)[0])
        outdegrees.append(degree(d, 0)[0])

    x = np.array(x)
    e = np.array(e)
    g = np.array(g)
    indegrees = np.array(indegrees)
    outdegrees = np.array(outdegrees)

    info = {
        "graphs": {
            "sum": g.sum(),
            "batch_mean": g.mean(),
            "batch_std": g.std(),
            "batch_indegree_mean": indegrees.mean(),
            "batch_outdegree_mean": outdegrees.mean(),
            "batch_indegree_std": indegrees.std(),
            "batch_outdegree_std": outdegrees.std(),
        },
        "edges": {"sum": e.sum(), "batch_mean": e.mean(), "batch_std": e.std()},
        "nodes": {"sum": x.sum(), "batch_mean": x.mean(), "batch_std": x.std()},
    }

    return info



from ._nx_apply_to_key import NetworkxTransformFeatureData
from caldera.utils import functional as fn


def filter_dict(keys):
    if keys is None:
        return None
    return fn.compose(lambda x: x.items(), fn.filter_each(lambda k: k[0] in keys), dict)


class NetworkxFilterDataKeys(NetworkxTransformFeatureData):
    def __init__(self, node_keys=None, edge_keys=None, global_keys=None):
        self.node_keys = node_keys
        self.edge_keys = edge_keys
        self.global_keys = global_keys

        node_transform = filter_dict(node_keys)
        edge_transform = filter_dict(edge_keys)
        glob_transform = filter_dict(global_keys)

        super().__init__(
            node_transform=node_transform,
            edge_transform=edge_transform,
            global_transform=glob_transform,
        )
